from typing import Optional
from .base import (SupportsMoveToRecordsDirectory,
                   SupportsToDataframesSource)
from ...db.quoting import quote_schema_and_table
from ...db import DBDriver
from ..records_directory import RecordsDirectory
from ..processing_instructions import ProcessingInstructions
from ..records_format import BaseRecordsFormat
from ..unload_plan import RecordsUnloadPlan
from ..results import MoveResult
from sqlalchemy.engine import Engine
from contextlib import contextmanager
from ..schema import RecordsSchema
from ...url.resolver import UrlResolver
import logging
from typing import Iterator, List, TYPE_CHECKING
if TYPE_CHECKING:
    from .dataframes import DataframesRecordsSource  # noqa
    from pandas import DataFrame  # noqa

logger = logging.getLogger(__name__)


class TableRecordsSource(SupportsMoveToRecordsDirectory,
                         SupportsToDataframesSource):
    def __init__(self,
                 schema_name: str,
                 table_name: str,
                 driver: DBDriver,
                 url_resolver: UrlResolver) -> None:
        self.schema_name = schema_name
        self.table_name = table_name
        self.driver = driver
        self.records_format = driver.best_records_format()
        self.url_resolver = url_resolver

    def known_supported_records_formats(self) -> List[BaseRecordsFormat]:
        return self.driver.known_supported_records_formats_for_unload()

    def can_move_to_this_format(self,
                                target_records_format: BaseRecordsFormat) -> bool:
        return self.driver.can_unload_this_format(target_records_format)

    @contextmanager
    def to_dataframes_source(self,
                             processing_instructions: ProcessingInstructions) -> \
            Iterator['DataframesRecordsSource']:
        from .dataframes import DataframesRecordsSource  # noqa
        import pandas

        db = self.driver.db
        records_schema = self.pull_records_schema()

        columns = db.dialect.get_columns(db,
                                         self.table_name,
                                         schema=self.schema_name)

        num_columns = len(columns)
        if num_columns == 0:
            raise LookupError(f"Could not find {self.schema_name}.{self.table_name}")
        entries_per_chunk = 2000000
        chunksize = int(entries_per_chunk / num_columns)
        logger.info(f"Exporting in chunks of up to {chunksize} rows by {num_columns} columns")

        quoted_table = quote_schema_and_table(db, self.schema_name, self.table_name)
        chunks: Iterator['DataFrame'] = \
            pandas.read_sql(f"SELECT * FROM {quoted_table}",
                            con=db,
                            chunksize=chunksize)
        yield DataframesRecordsSource(dfs=self.with_cast_dataframe_types(records_schema, chunks),
                                      records_schema=records_schema,
                                      processing_instructions=processing_instructions)

    def with_cast_dataframe_types(self,
                                  records_schema,
                                  dfs: Iterator['DataFrame']) -> Iterator['DataFrame']:
        for df in dfs:
            yield records_schema.cast_dataframe_types(df)

    def pull_records_schema(self) -> RecordsSchema:
        return RecordsSchema.from_db_table(self.schema_name, self.table_name,
                                           driver=self.driver)

    def move_to_records_directory(self,
                                  records_directory: RecordsDirectory,
                                  records_format: BaseRecordsFormat,
                                  processing_instructions: ProcessingInstructions,
                                  schema_name: Optional[str]=None,
                                  table_name: Optional[str]=None,
                                  engine: Optional[Engine]=None) -> MoveResult:
        unload_plan = RecordsUnloadPlan(records_format=records_format,
                                        processing_instructions=processing_instructions)
        export_count = self.driver.unload(schema=self.schema_name, table=self.table_name,
                                          unload_plan=unload_plan,
                                          directory=records_directory)
        records_schema = self.pull_records_schema()
        records_directory.save_format(unload_plan.records_format)
        records_directory.save_schema(records_schema)
        records_directory.finalize_manifest()

        return MoveResult(move_count=export_count, output_urls=None)

    def __str__(self) -> str:
        return f"{type(self).__name__}({self.driver.db_engine.name})"



from records_mover.records.targets.base import (
    SupportsMoveFromRecordsDirectory,
    SupportsMoveFromTempLocAfterFillingIt,
    MightSupportMoveFromFileobjsSource,
    SupportsMoveFromDataframes,
)
from sqlalchemy.engine import Engine, Connection
from records_mover.records.prep import TablePrep, TargetTableDetails
from records_mover.records.records_format import BaseRecordsFormat
from records_mover.db import DBDriver
from records_mover.records.records_directory import RecordsDirectory
from records_mover.records.processing_instructions import ProcessingInstructions
from records_mover.records.existing_table_handling import ExistingTableHandling
from records_mover.records.results import MoveResult
from records_mover.records.sources import SupportsMoveToRecordsDirectory
from records_mover.records.sources.fileobjs import FileobjsSource
from records_mover.records.targets.table.move_from_records_directory import (
    DoMoveFromRecordsDirectory
)
from records_mover.records.targets.table.move_from_fileobjs_source import DoMoveFromFileobjsSource
from records_mover.records.targets.table.move_from_temp_loc_after_filling_it import (
    DoMoveFromTempLocAfterFillingIt
)
import logging
from typing import Callable, Union, Optional, Dict, List, TYPE_CHECKING
if TYPE_CHECKING:
    from records_mover.records.sources.dataframes import DataframesRecordsSource


logger = logging.getLogger(__name__)


class TableRecordsTarget(SupportsMoveFromRecordsDirectory,
                         SupportsMoveFromTempLocAfterFillingIt,
                         MightSupportMoveFromFileobjsSource,
                         SupportsMoveFromDataframes,
                         TargetTableDetails):
    def __init__(self,
                 schema_name: str,
                 table_name: str,
                 db_engine: Engine,
                 db_driver: Callable[[Union[Engine, Connection]], DBDriver],
                 add_user_perms_for: Optional[Dict[str, List[str]]] = None,
                 add_group_perms_for: Optional[Dict[str, List[str]]] = None,
                 existing_table_handling: ExistingTableHandling =
                 ExistingTableHandling.DELETE_AND_OVERWRITE,
                 drop_and_recreate_on_load_error: bool = False) -> None:
        self.schema_name = schema_name
        self.table_name = table_name
        self.db_driver = db_driver  # type: ignore
        self.db_engine = db_engine
        self.add_user_perms_for = add_user_perms_for
        self.add_group_perms_for = add_group_perms_for
        self.existing_table_handling = existing_table_handling
        self.drop_and_recreate_on_load_error = drop_and_recreate_on_load_error
        self.prep = TablePrep(self)
        # advertise what format we prefer to be given for mover paths
        # that don't yet support full records negotiation.
        #
        # https://app.asana.com/0/1128138765527694/1130105834872106
        self.records_format = next(iter(self.known_supported_records_formats()), None)

    def move_from_records_directory(self,
                                    directory: RecordsDirectory,
                                    processing_instructions: ProcessingInstructions,
                                    override_records_format: Optional[BaseRecordsFormat] = None)\
            -> MoveResult:
        return DoMoveFromRecordsDirectory(self.prep,
                                          self,
                                          directory,
                                          processing_instructions,
                                          override_records_format).move()

    def move_from_fileobjs_source(self,
                                  fileobjs_source: FileobjsSource,
                                  processing_instructions: ProcessingInstructions) -> MoveResult:
        return DoMoveFromFileobjsSource(self.prep,
                                        self,
                                        fileobjs_source,
                                        processing_instructions).move()

    def can_move_from_fileobjs_source(self) -> bool:
        driver = self.db_driver(self.db_engine)
        return driver.can_load_from_fileobjs()

    def can_load_direct(self, scheme: str) -> bool:
        driver = self.db_driver(self.db_engine)
        return driver.best_scheme_to_load_from() == scheme

    def known_supported_records_formats(self) -> List[BaseRecordsFormat]:
        driver = self.db_driver(self.db_engine)
        return driver.known_supported_records_formats_for_load()

    def can_move_from_this_format(self,
                                  source_records_format: BaseRecordsFormat) -> bool:
        """Return true if writing the specified format satisfies our format
        needs"""
        driver = self.db_driver(self.db_engine)
        return driver.can_load_this_format(source_records_format)

    def move_from_temp_loc_after_filling_it(self,
                                            records_source:
                                            SupportsMoveToRecordsDirectory,
                                            processing_instructions: ProcessingInstructions)\
            -> MoveResult:
        return DoMoveFromTempLocAfterFillingIt(self.prep,
                                               self,
                                               self,
                                               records_source,
                                               processing_instructions).move()

    def move_from_dataframes_source(self,
                                    dfs_source: 'DataframesRecordsSource',
                                    processing_instructions:
                                    ProcessingInstructions) -> MoveResult:
        from records_mover.records.targets.table.move_from_dataframes_source import (
            DoMoveFromDataframesSource
        )

        return DoMoveFromDataframesSource(self.prep,
                                          self,
                                          self,
                                          dfs_source,
                                          processing_instructions).move()

    def __str__(self) -> str:
        return f"{type(self).__name__}({self.db_engine.name})"



from airflow.contrib.hooks.gcp_api_base_hook import GoogleCloudBaseHook
from typing import Iterable, Optional, TYPE_CHECKING
if TYPE_CHECKING:
    # see the 'gsheets' extras_require option in setup.py - needed for this!
    import google.auth.credentials  # noqa


class GoogleCloudCredentialsHook(GoogleCloudBaseHook):
    def get_conn(self) -> 'google.auth.credentials.Credentials':
        return self._get_credentials()

    def scopes(self) -> Iterable[str]:
        scope: Optional[str] = self._get_field('scope', None)
        scopes: Iterable[str]
        if scope is not None:
            scopes = [s.strip() for s in scope.split(',')]
        else:
            # https://github.com/apache/incubator-airflow/blob/master/airflow/contrib/hooks/gcp_api_base_hook.py#L32
            scopes = ('https://www.googleapis.com/auth/cloud-platform',)
        return scopes



from typing import Callable, Optional
from records_mover.records.results import MoveResult
from records_mover.db import DBDriver
from records_mover.records.prep import TablePrep
from records_mover.records.table import TargetTableDetails
from records_mover.records.existing_table_handling import ExistingTableHandling
import logging

logger = logging.getLogger(__name__)


def prep_and_load(tbl: TargetTableDetails,
                  prep: TablePrep,
                  schema_sql: str,
                  load: Callable[[DBDriver], Optional[int]],
                  reset_before_reload: Callable[[], None] = lambda: None) -> MoveResult:
    logger.info(f"Connecting to database...")
    with tbl.db_engine.begin() as db:
        driver = tbl.db_driver(db)
        prep.prep(schema_sql=schema_sql, driver=driver)
    with tbl.db_engine.begin() as db:
        # This second transaction ensures the table has been created
        # before non-transactional statements like Redshift's COPY
        # take place.  Otherwise you'll get an error like:
        #
        #  Cannot COPY into nonexistent table
        driver = tbl.db_driver(db)
        try:
            import_count = load(driver)
        except driver.load_failure_exception():
            if not tbl.drop_and_recreate_on_load_error:
                raise
            reset_before_reload()
            with tbl.db_engine.begin() as db:
                driver = tbl.db_driver(db)
                prep.prep(schema_sql=schema_sql,
                          driver=driver,
                          existing_table_handling=ExistingTableHandling.DROP_AND_RECREATE)
                import_count = load(driver)
        return MoveResult(move_count=import_count, output_urls=None)



"""Defines hints supported by the job config parser."""
from ...utils.json_schema import JsonParameter, JsonSchemaDocument
from typing import Optional


class SupportedHint:
    """Definition for supported hints"""

    def __init__(self,
                 schema: JsonParameter,
                 target_hint_name: Optional[str] = None):
        self.schema = schema
        self.target_hint_name = target_hint_name or schema.name

    @property
    def config_name(self) -> str:
        return self.schema.name


QUOTING_DESCRIPTION =\
        ('How quotes are applied to individual fields. '
         'all: quote all fields. '
         'minimal: quote only fields that contain ambiguous characters (the '
         'delimiter, the escape character, or a line terminator). '
         'default: never quote fields.')

SUPPORTED_HINTS = [
    SupportedHint(
        JsonParameter('field-delimiter',
                      JsonSchemaDocument('string',
                                         description=('Character used between fields '
                                                      '(default is comma)')),
                      optional=True)),
    SupportedHint(
        JsonParameter('compression',
                      JsonSchemaDocument(['string', 'null'],
                                         enum=['BZIP', 'GZIP', 'LZO', None],
                                         description='Compression type of the file.'),
                      optional=True)),
    SupportedHint(
        JsonParameter('escape',
                      JsonSchemaDocument(['string', 'null'],
                                         enum=['\\', None],
                                         description="Character used to escape strings"),
                      optional=True)),
    SupportedHint(
        JsonParameter('quoting',
                      JsonSchemaDocument(['string', 'null'],
                                         enum=['all', 'minimal', 'nonnumeric', None],
                                         description=QUOTING_DESCRIPTION),
                      optional=True)),
    SupportedHint(
        JsonParameter('encoding',
                      JsonSchemaDocument(['string'],
                                         enum=[
                                             'UTF8', 'UTF16', 'UTF16LE', 'UTF16BE',
                                             'LATIN1', 'CP1252'
                                         ],
                                         description="Text encoding of file"),
                      optional=True)),
]
SUPPORTED_HINT_LOOKUP = {hint.config_name: hint for hint in SUPPORTED_HINTS}



from records_mover.db.quoting import quote_schema_and_table, quote_value
from typing import Optional


def vertica_import_sql(schema: str,
                       table: str,
                       db_engine,
                       gzip: bool,
                       delimiter: str,
                       trailing_nullcols: bool,
                       record_terminator: str,
                       null_as: str,
                       rejectmax: int,
                       enforcelength: bool,
                       error_tolerance: bool,
                       abort_on_error: bool,
                       load_method: str,
                       no_commit: bool,
                       escape_as: Optional[str]=None,
                       enclosed_by: Optional[str]=None,
                       stream_name: Optional[str]=None,
                       skip=0,
                       rejected_data_table: Optional[str]=None,
                       rejected_data_schema: Optional[str]=None) -> str:

    # https://my.vertica.com/docs/8.1.x/HTML/index.htm#Authoring/SQLReferenceManual/Statements/COPY/COPY.htm
    # https://my.vertica.com/docs/8.1.x/HTML/index.htm#Authoring/SQLReferenceManual/Statements/COPY/Parameters.htm
    # https://my.vertica.com/docs/8.1.x/HTML/index.htm#Authoring/AdministratorsGuide/BulkLoadCOPY/SpecifyingCOPYFROMOptions.htm
    import_sql_template = """
            COPY {schema_and_table}
            FROM STDIN
            {gzip}
            {delimiter_as}
            {trailing_nullcols}
            {null_as}
            {escape}
            {enclosed_by}
            {record_terminator}
            {skip}
            {skip_bytes}
            {trim}
            {rejectmax}
            {rejected_data}
            {exceptions}
            {enforcelength}
            {error_tolerance}
            {abort_on_error}
            {storage}
            {stream_name}
            {no_commit}
            ;
    """

    def quote(value: str) -> str:
        return quote_value(db_engine, value)

    if rejected_data_table is not None and rejected_data_schema is not None:
        rejected_target = quote_schema_and_table(db_engine,
                                                 rejected_data_schema, rejected_data_table)
        rejected_data = f"REJECTED DATA AS TABLE {rejected_target}"
    else:
        rejected_data = ''

    import_sql = import_sql_template.format(
        schema_and_table=quote_schema_and_table(db_engine, schema, table),
        # https://forum.vertica.com/discussion/238556/reading-gzip-files-from-s3-into-vertica
        gzip='GZIP' if gzip else '',
        # The capital E in the next line specifies a string literal
        # that will be backslash-interpreted; the call to python's
        # "repr()" built-in will encode the string with backslashes.
        delimiter_as=f"DELIMITER AS E{repr(delimiter)}",
        trailing_nullcols="TRAILING NULLCOLS" if trailing_nullcols else '',
        null_as=f"NULL AS {quote(null_as)}" if null_as is not None else '',
        escape=f"ESCAPE AS E{repr(escape_as)}" if escape_as is not None else "NO ESCAPE",
        enclosed_by=f"ENCLOSED BY E{repr(enclosed_by)}" if enclosed_by is not None else '',
        record_terminator=f"RECORD TERMINATOR E{repr(record_terminator)}",
        skip=f"SKIP {skip}",
        skip_bytes='',  # only for fixed-width
        trim="",  # only for fixed-width
        rejectmax=f"REJECTMAX {rejectmax}" if rejectmax is not None else '',
        rejected_data=rejected_data,
        exceptions='',   # not yet supported
        enforcelength='ENFORCELENGTH' if enforcelength else '',
        error_tolerance='ERROR TOLERANCE' if error_tolerance else '',
        abort_on_error='ABORT ON ERROR' if abort_on_error else '',
        storage=f"STORAGE {load_method}",
        stream_name=f"STREAM NAME {quote(stream_name)}" if stream_name is not None else '',
        no_commit="NO COMMIT" if no_commit else '',
    )

    return import_sql



from .vertica.vertica_db_driver import VerticaDBDriver
from .redshift.redshift_db_driver import RedshiftDBDriver
from .bigquery.bigquery_db_driver import BigQueryDBDriver
from .postgres.postgres_db_driver import PostgresDBDriver
from .driver import DBDriver
import sqlalchemy
from typing import Union


def db_driver(db: Union[sqlalchemy.engine.Engine,
                        sqlalchemy.engine.Connection],
              **kwargs) -> DBDriver:
    engine: sqlalchemy.engine.Engine = db.engine

    if engine.name == 'vertica':
        return VerticaDBDriver(db, **kwargs)
    elif engine.name == 'redshift':
        return RedshiftDBDriver(db, **kwargs)
    elif engine.name == 'bigquery':
        return BigQueryDBDriver(db, **kwargs)
    elif engine.name == 'postgresql':
        return PostgresDBDriver(db, **kwargs)
    else:
        return DBDriver(db, **kwargs)



from .base import BaseFileUrl, BaseDirectoryUrl
from . import FileUrl, DirectoryUrl


class UrlResolver:
    def __init__(self, **kwargs) -> None:
        self.kwargs = kwargs

    def file_url(self, url: str) -> BaseFileUrl:
        return FileUrl(url, **self.kwargs)

    def directory_url(self, url: str) -> BaseDirectoryUrl:
        return DirectoryUrl(url, **self.kwargs)



from .processing_instructions import ProcessingInstructions
from .records_format import BaseRecordsFormat


class RecordsLoadPlan:
    def __init__(self,
                 processing_instructions: ProcessingInstructions,
                 records_format: BaseRecordsFormat) -> None:
        self.records_format = records_format
        self.processing_instructions = processing_instructions



import boto3
from db_facts.db_facts_types import DBFacts
from typing import TYPE_CHECKING, Iterable
if TYPE_CHECKING:
    # see the 'gsheets' extras_require option in setup.py - needed for this!
    import google.auth.credentials  # noqa


# this interfaces here are probably unstable until we figure out how
# best to integrate airflow hooks and connections (and maybe
# Kubernetes secrets) in with it.  The current (unfinalized and
# untested) thought is to use the *_creds_name argument to reflect the
# Airflow connection name, so a generic job can call
# session.creds.cred_type(some_name_from_its_arguments) and get a
# valid cred both on the command line and in Airflow.
#
# If this seems to work, we can extend this idea out to Kubernetes
# secrets by writing a separate backend for it (and if we need to
# support both Airflow and Kubernetes secrets depending on the
# situation, let the job context pick which backend we're using for
# creds based on out of band information, like how it was constructed
# or environment variables).
class BaseCreds():
    def google_sheets(self, gcp_creds_name: str) -> 'google.auth.credentials.Credentials':
        scopes = ('https://www.googleapis.com/auth/spreadsheets',)
        return self._gcp_creds(gcp_creds_name, scopes)

    def _gcp_creds(self, gcp_creds_name: str,
                   scopes: Iterable[str]) -> 'google.auth.credentials.Credentials':
        raise NotImplementedError

    def db_facts(self, db_creds_name: str) -> DBFacts:
        raise NotImplementedError

    def boto3_session(self, aws_creds_name: str) -> boto3.session.Session:
        raise NotImplementedError



from abc import ABCMeta, abstractmethod
import json
from typing import Mapping, Optional, TYPE_CHECKING
if TYPE_CHECKING:
    from . import RecordsFormatType  # noqa


class BaseRecordsFormat(metaclass=ABCMeta):
    format_type: 'RecordsFormatType'  # noqa

    def json(self) -> str:
        contents = ''
        contents_data = self.config()
        if contents_data is not None:
            contents = json.dumps(contents_data)
        return contents

    def config(self) -> Optional[Mapping[str, object]]:
        return None

    @abstractmethod
    def generate_filename(self, basename: str) -> str:
        pass



import numpy
from typing import Optional, Tuple, TYPE_CHECKING
from .constraints import RecordsSchemaFieldConstraints
if TYPE_CHECKING:
    from .types import FieldType  # noqa


def details_from_numpy_dtype(dtype: numpy.dtype,
                             unique: bool) -> Tuple['FieldType',
                                                    RecordsSchemaFieldConstraints]:
    from ..field import RecordsSchemaField
    basename = dtype.base.name
    field_type: Optional['FieldType']
    if basename.startswith('datetime64'):
        has_tz = getattr(dtype, "tz", None) is not None
        if basename == 'datetime64[D]':
            field_type = 'date'
        else:
            if has_tz:
                # See: 'Represent pandas datetime64 with timezone in records schema'
                #
                # https://app.asana.com/0/53283930106309/1132706099772565
                field_type = 'datetimetz'
            else:
                field_type = 'datetime'
    else:
        field_type = RecordsSchemaField.python_type_to_field_type(dtype.type)
        if field_type is None:
            raise NotImplementedError(f"Teach me how to handle Pandas/numpy dtype {dtype}"
                                      f"which is dtype.type {dtype.type}")

    constraints = RecordsSchemaFieldConstraints.from_numpy_dtype(dtype, unique=unique)

    return (field_type, constraints)



from typing import TYPE_CHECKING
# Be sure to add new things below in FieldType, too
RECORDS_FIELD_TYPES = {
    'integer',
    'decimal',
    'string',
    'boolean',
    'date',
    'time',
    'timetz',
    'datetime',
    'datetimetz'
}

if TYPE_CHECKING:
    from typing_extensions import Literal  # noqa
    FieldType = Literal['integer',
                        'decimal',
                        'string',
                        'boolean',
                        'date',
                        'time',
                        'timetz',
                        'datetime',
                        'datetimetz']



from ...utils.json_schema import method_signature_to_json_schema, JsonParameter, JsonSchemaDocument
from ..existing_table_handling import ExistingTableHandling
from typing import Any, Dict, List, Callable
from ...types import JsonSchema
from .hints import SUPPORTED_HINTS


def method_to_json_schema(method: Callable[..., Any]) -> JsonSchema:
    special_handling: Dict[str, List[JsonParameter]] = {
        'google_cloud_creds': [JsonParameter('gcp_creds_name', JsonSchemaDocument('string'))],
        'db_engine': [JsonParameter('db_name', JsonSchemaDocument('string'))],
        'records_format': ([JsonParameter('variant', JsonSchemaDocument('string'), optional=True)] +
                           [hint.schema for hint in SUPPORTED_HINTS]),
        'initial_hints': [hint.schema for hint in SUPPORTED_HINTS],
        'existing_table_handling':
        [JsonParameter('existing_table',
                       JsonSchemaDocument('string',
                                          enum=[k.lower()
                                                for k in ExistingTableHandling.__members__],
                                          default='delete_and_overwrite'),
                       optional=True)],
    }

    parameters_to_ignore = [
        'self',
        # maybe in the future we can have people point us to a JSON
        # file for hints...
        'hints',
        # ...and maybe this particular way of passing in dicts:
        'add_user_perms_for',
        'add_group_perms_for',
        # oh yeah, and a way to pass a filename in for those that have
        # hand-crafted a records schema file yet don't have a full
        # records directory...
        'records_schema'
    ]

    return method_signature_to_json_schema(method,
                                           special_handling=special_handling,
                                           parameters_to_ignore=parameters_to_ignore)



from ...utils.retry import google_sheets_retry
from .base import SupportsMoveFromDataframes
import math
import google.auth.credentials
import numpy as np
from ..results import MoveResult
from ..processing_instructions import ProcessingInstructions
from apiclient.discovery import build
from googleapiclient.errors import HttpError
import google_auth_httplib2
import httplib2
from typing import Any, Union, List
from ..sources.dataframes import DataframesRecordsSource
import logging

logger = logging.getLogger(__name__)

SheetsService = Any
CellData = Union[int, str, float]


class GoogleSheetsRecordsTarget(SupportsMoveFromDataframes):
    def __init__(self,
                 spreadsheet_id: str,
                 sheet_name: str,
                 google_cloud_creds: google.auth.credentials.Credentials) -> None:
        self.spreadsheet_id = spreadsheet_id
        self.sheet_name = sheet_name
        self.google_cloud_creds = google_cloud_creds

    @google_sheets_retry()
    def _create_sheet(self,
                      service: SheetsService,
                      spreadsheet_id: str,
                      sheet_name: str) -> None:
        # https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/batchUpdate
        # https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/request#AddSheetRequest
        # https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/sheets#SheetProperties
        batch_update_spreadsheet_request_body = {
            'requests': [
                {
                    "addSheet": {
                        "properties": {
                            "title": sheet_name
                        }
                    }
                }
            ],
        }

        request = service.spreadsheets().batchUpdate(spreadsheetId=spreadsheet_id,
                                                     body=batch_update_spreadsheet_request_body)
        request.execute()

    @google_sheets_retry()
    def _clear_sheet(self,
                     service: SheetsService,
                     spreadsheet_id: str,
                     sheet_name: str) -> None:
        service.spreadsheets().values().clear(
            spreadsheetId=spreadsheet_id,
            range=sheet_name,
            body={}
        ).execute()

    def _clear_or_create_sheet(self,
                               service: SheetsService,
                               spreadsheet_id: str,
                               sheet_name: str) -> None:
        try:
            self._clear_sheet(service, spreadsheet_id, sheet_name)
        except HttpError:
            self._create_sheet(service, spreadsheet_id, sheet_name)

    @google_sheets_retry()
    def _naive_load_to_sheet(self,
                             service: SheetsService,
                             sheet_id: str,
                             sheet_name: str,
                             data: List[List[CellData]]) -> None:
        # assumes an empty sheet; starts in the upper left corner, cell A1
        body = {
            'values': data
        }

        service.spreadsheets().values().update(
            spreadsheetId=sheet_id,
            range=sheet_name,
            valueInputOption='RAW',
            body=body,
        ).execute()

    def _authorize(self) -> google_auth_httplib2.AuthorizedHttp:
        credentials = self.google_cloud_creds
        http = httplib2.Http()
        authed_http = google_auth_httplib2.AuthorizedHttp(
            credentials, http=http)
        return authed_http

    def _get_service(self) -> SheetsService:
        http_authorized = self._authorize()
        return build('sheets', 'v4', http=http_authorized)

    def as_json_serializable(self, cell: Any) -> Any:
        if isinstance(cell, np.generic):
            native = np.asscalar(cell)
        else:
            native = cell
        if isinstance(cell, float) and math.isnan(native):
            native = 'NaN'
        return native

    def move_from_dataframes_source(self,
                                    dfs_source: DataframesRecordsSource,
                                    processing_instructions:
                                    ProcessingInstructions) -> MoveResult:
        dfs_iterator = iter(dfs_source.dfs)
        df = next(dfs_iterator)
        if next(dfs_iterator, None) is not None:
            raise NotImplementedError("Please teach me how to handle "
                                      "more than one dataframe at a time")
        data = df.to_records(index=dfs_source.include_index)
        # these need to be regular Python types so that we can
        # serialize them to JSON over the network as we send them to
        # Google Sheets.
        json_encodable_data = [[self.as_json_serializable(cell) for cell in row] for row in data]
        service = self._get_service()
        self._clear_or_create_sheet(service, self.spreadsheet_id, self.sheet_name)
        self._naive_load_to_sheet(service, self.spreadsheet_id, self.sheet_name,
                                  json_encodable_data)
        return MoveResult(move_count=len(data), output_urls=None)



from abc import ABCMeta
from sqlalchemy.engine import Engine, Connection
from typing import Union, Optional, Dict, List
from records_mover.records.existing_table_handling import ExistingTableHandling
from records_mover.db import DBDriver
import logging

logger = logging.getLogger(__name__)


class TargetTableDetails(metaclass=ABCMeta):
    schema_name: str
    table_name: str
    db_engine: Engine
    add_user_perms_for: Optional[Dict[str, List[str]]]
    add_group_perms_for: Optional[Dict[str, List[str]]]
    existing_table_handling: ExistingTableHandling
    drop_and_recreate_on_load_error: bool

    # This should really be a like the above - but mypy gets confused
    # by a function being assigned as a field.
    #
    # Unfortunately it also gets confused by @abstractmethod.
    #
    # https://github.com/python/mypy/issues/5485
    # @abstractmethod
    def db_driver(self, db: Union[Engine, Connection]) -> DBDriver:
        ...



import logging
import json
from .sqlalchemy import schema_from_db_table, schema_to_schema_sql
from typing import List, Dict, Mapping, IO, Any, TYPE_CHECKING
from ..field import RecordsSchemaField
from ...records_format import BaseRecordsFormat
from ...processing_instructions import ProcessingInstructions
from .known_representation import RecordsSchemaKnownRepresentation
from ..errors import UnsupportedSchemaError
if TYPE_CHECKING:
    from pandas import DataFrame

    from ....db import DBDriver  # noqa
    from typing_extensions import Literal

    from mypy_extensions import TypedDict

    from ..field import FieldDict
    from .known_representation import KnownRepresentationDict

    class MandatoryRecordsSchemaDict(TypedDict):
        schema: Literal['bltypes/v1']
        fields: Dict[str, FieldDict]

    class RecordsSchemaDict(MandatoryRecordsSchemaDict, total=False):
        known_representations: Dict[str, KnownRepresentationDict]


logger = logging.getLogger(__name__)


class RecordsSchema:
    def __init__(self,
                 fields: List[RecordsSchemaField],
                 known_representations: Mapping[str, RecordsSchemaKnownRepresentation]) -> None:
        self.fields = fields
        self.known_representations = known_representations

    def __str__(self) -> str:
        types = {field.name: field.field_type for field in self.fields}
        return f"RecordsSchema(types={types})"

    @staticmethod
    def from_db_table(schema_name: str, table_name: str, driver: 'DBDriver') -> 'RecordsSchema':
        return schema_from_db_table(schema_name=schema_name,
                                    table_name=table_name,
                                    driver=driver)

    def to_data(self) -> 'RecordsSchemaDict':
        data: 'RecordsSchemaDict' = {
            'schema': 'bltypes/v1',
            'fields': {},
        }
        i = 1
        for field in self.fields:
            data['fields'][field.name] = field.to_data()
            data['fields'][field.name]['index'] = i
            i = i + 1
        if self.known_representations:
            data['known_representations'] = {
                system: known_representation.to_data()
                for system, known_representation
                in self.known_representations.items()
            }
        return data

    def to_json(self) -> str:
        return json.dumps(self.to_data())

    @staticmethod
    def from_data(data: 'RecordsSchemaDict') -> 'RecordsSchema':
        schema_ver = data["schema"]
        if schema_ver != "bltypes/v1":
            raise UnsupportedSchemaError(
                f"unsupported record specification schema {schema_ver}")
        known_representations = {}
        for name, known_rep_data in data.get("known_representations", {}).items():
            known_representation = RecordsSchemaKnownRepresentation.from_data(known_rep_data)
            if known_representation is not None:
                known_representations[name] = known_representation
        return RecordsSchema(
            fields=[
                RecordsSchemaField.from_data(name, field_data)
                for name, field_data
                in data.get("fields", {}).items()
            ],
            known_representations=known_representations)

    @staticmethod
    def from_json(json_str: str) -> 'RecordsSchema':
        data = json.loads(json_str)
        return RecordsSchema.from_data(data)

    def to_schema_sql(self,
                      driver: 'DBDriver',
                      schema_name: str,
                      table_name: str) -> str:
        return schema_to_schema_sql(records_schema=self,
                                    driver=driver,
                                    schema_name=schema_name,
                                    table_name=table_name)

    @staticmethod
    def from_fileobjs(fileobjs: List[IO[bytes]],
                      records_format: BaseRecordsFormat,
                      processing_instructions: ProcessingInstructions) -> 'RecordsSchema':
        from records_mover.records.csv_streamer import stream_csv
        from records_mover.pandas import purge_unnamed_unused_columns

        if len(fileobjs) != 1:
            # https://app.asana.com/0/53283930106309/1131698268455054
            raise NotImplementedError('Cannot currently sniff schema from mulitple '
                                      'files--please provide explicit schema JSON')
        fileobj = fileobjs[0]
        if not fileobj.seekable():
            raise NotImplementedError('Cannot currently sniff schema from a pure stream--'
                                      'please save file to disk and load from there or '
                                      'provide explicit schema JSON')
        with stream_csv(fileobj, records_format.hints) as reader:  # type: ignore
            # Parse schema from sample df

            sample_row_count = processing_instructions.max_inference_rows
            if sample_row_count is not None:
                df = reader.get_chunk(sample_row_count)
            else:
                df = reader.read()

            fileobj.seek(0)

            df = purge_unnamed_unused_columns(df)
            schema = RecordsSchema.from_dataframe(df, processing_instructions,
                                                  include_index=False)

            schema.refine_from_dataframe(df,
                                         processing_instructions=processing_instructions)
            return schema

    def refine_from_dataframe(self,
                              df: 'DataFrame',
                              processing_instructions:
                              ProcessingInstructions = ProcessingInstructions()) -> None:
        """
        Adjust records schema based on facts found from a dataframe.
        """
        from .pandas import refine_schema_from_dataframe
        return refine_schema_from_dataframe(records_schema=self,
                                            df=df,
                                            processing_instructions=processing_instructions)

    def cast_dataframe_types(self,
                             df: 'DataFrame') -> 'DataFrame':
        """
        Returns a new dataframe with types that match what we know from this records schema.
        """
        col_mappings = {field.name: field.to_numpy_dtype() for field in self.fields}
        if len(col_mappings) == 0:
            # .as_type() doesn't like being given an empty map!
            return df
        return df.astype(col_mappings)

    def assign_dataframe_names(self,
                               include_index: bool,
                               df: 'DataFrame') -> 'DataFrame':
        """Returns a new dataframe with index/series names that match what we
        know from this records schema.  Useful when we've created a
        dataframe from a file where the headers don't match the exact
        names or when the names aren't provided
        """
        df_names = [label for label in df]
        field_names = [field.name for field in self.fields]
        index_mapping = None
        if include_index:
            index_name = field_names[0]
            # RecordsMover supports only single indexes for the moment.
            #
            # https://app.asana.com/0/1128138765527694/1161071033650873
            index_mapping = {0: index_name}

            # remove first name, which applies to index
            field_names = field_names[1:]
        if len(df_names) != len(field_names):
            raise SyntaxError("Number of DataFrame names didn't match number of schema names!"
                              f"DataFrame: {df_names}.  Schema names: {field_names}")
        name_mapping = dict(zip(df_names, field_names))
        df = df.rename(columns=name_mapping,
                       copy=False,
                       errors='raise')
        if include_index:
            return df.rename(index=index_mapping,
                             copy=False,
                             errors='raise')
        else:
            return df

    @staticmethod
    def from_dataframe(df: 'DataFrame',
                       processing_instructions: ProcessingInstructions,
                       include_index: bool) -> 'RecordsSchema':
        from .pandas import schema_from_dataframe
        return schema_from_dataframe(df=df,
                                     processing_instructions=processing_instructions,
                                     include_index=include_index)

    def to_empty_dataframe(self) -> 'DataFrame':
        from pandas import DataFrame
        df_data: Dict[str, List[Any]] = {field.name: [] for field in self.fields}
        raw_df = DataFrame(df_data)
        return self.cast_dataframe_types(raw_df)

    def convert_datetimes_to_datetimetz(self) -> 'RecordsSchema':
        return RecordsSchema(fields=[field.convert_datetime_to_datetimetz()
                                     for field in self.fields],
                             known_representations=self.known_representations)



import itertools
from .base import SupportsToFileobjsSource
from ..processing_instructions import ProcessingInstructions
from ..pandas import pandas_to_csv_options
from ..schema import RecordsSchema
from contextlib import contextmanager
from ..records_format import BaseRecordsFormat, DelimitedRecordsFormat, ParquetRecordsFormat
from .fileobjs import FileobjsSource  # noqa
from tempfile import NamedTemporaryFile
from ..hints import complain_on_unhandled_hints
import logging
from typing import Iterator, Iterable, Optional, Union, Dict, IO, Callable, TYPE_CHECKING
if TYPE_CHECKING:
    from pandas import DataFrame


logger = logging.getLogger(__name__)


class DataframesRecordsSource(SupportsToFileobjsSource):
    def __init__(self,
                 dfs: Iterable['DataFrame'],
                 processing_instructions: ProcessingInstructions=ProcessingInstructions(),
                 records_schema: Optional[RecordsSchema]=None,
                 include_index: bool=False) -> None:
        self.dfs = dfs
        self.processing_instructions = processing_instructions
        self.records_schema = records_schema
        self.include_index = include_index

    def pick_best_records_format(self,
                                 records_format_if_possible: Optional[BaseRecordsFormat]) ->\
            Union[ParquetRecordsFormat, DelimitedRecordsFormat]:
        # Ensure we don't accept a records format that we don't yet
        # have code to export to.
        if records_format_if_possible is None:
            return DelimitedRecordsFormat()
        if isinstance(records_format_if_possible, DelimitedRecordsFormat):
            return records_format_if_possible
        if isinstance(records_format_if_possible, ParquetRecordsFormat):
            return records_format_if_possible

        return DelimitedRecordsFormat()

    def peek(self) -> 'DataFrame':
        # https://stackoverflow.com/a/2425347/9795956
        dfs_iter = iter(self.dfs)
        first_df = next(dfs_iter)
        self.dfs = itertools.chain([first_df], dfs_iter)
        return first_df

    def initial_records_schema(self,
                               processing_instructions: ProcessingInstructions) -> RecordsSchema:
        records_schema = self.records_schema
        if records_schema is None:
            # We don't have any schema to begin with - let's
            # start with one based on this df:
            records_schema = self.schema_from_df(self.peek(), processing_instructions)
        return records_schema

    def serialize_dfs(self,
                      processing_instructions: ProcessingInstructions,
                      records_format: BaseRecordsFormat,
                      save_df: Callable[['DataFrame', str], None])\
            -> Iterator[FileobjsSource]:
        records_schema = self.initial_records_schema(processing_instructions)

        target_names_to_input_fileobjs: Dict[str, IO[bytes]] = {}
        i = 1

        for df in self.dfs:
            with NamedTemporaryFile(prefix='mover_seralized_dataframe') as output_file:
                output_filename = output_file.name
                logger.info(f"Writing CSV file to {output_filename}")
                save_df(df, output_filename)
                short_filename = records_format.generate_filename('data{:0>3}'.format(i))
                target_names_to_input_fileobjs[short_filename] = open(output_filename, 'rb')
                # pad with leading zeros to three digits so these files sort when listed
                if i == 2 and self.records_schema is None:
                    # Records mover is not yet smart enough to
                    # coalease multiple records schemas from different
                    # chunks into a single schema, so the result of
                    # initial_records_schema was based on only the
                    # first chunk.
                    #
                    # https://app.asana.com/0/1128138765527694/1161074649542801
                    logger.warning("Only checking first chunk for type inference")
                i = i + 1

        try:
            yield FileobjsSource(target_names_to_input_fileobjs=target_names_to_input_fileobjs,
                                 records_schema=records_schema,
                                 records_format=records_format)
        finally:
            for filename, fileobj in target_names_to_input_fileobjs.items():
                if not fileobj.closed:
                    fileobj.close()

    def schema_from_df(self, df: 'DataFrame',
                       processing_instructions: ProcessingInstructions) -> RecordsSchema:
        records_schema = RecordsSchema.from_dataframe(df,
                                                      self.processing_instructions,
                                                      include_index=self.include_index)
        if (processing_instructions.max_inference_rows is None or
           processing_instructions.max_inference_rows > 0):
            #
            # If we were provided with a RecordsSchema, assume
            # that the user wants us to use that verbatim (or they
            # can call .refine_from_dataframe() themselves.
            # Otherwise, gather information to create an efficient
            # schema on the target of the move.
            #
            records_schema.refine_from_dataframe(df, processing_instructions)

        return records_schema

    @contextmanager
    def to_fileobjs_source(self,
                           processing_instructions: ProcessingInstructions,
                           records_format_if_possible: Optional[BaseRecordsFormat]=
                           None) -> Iterator[FileobjsSource]:
        records_format = self.pick_best_records_format(records_format_if_possible)
        if isinstance(records_format, DelimitedRecordsFormat):
            unhandled_hints = set(records_format.hints.keys())
            options = pandas_to_csv_options(records_format,
                                            unhandled_hints,
                                            self.processing_instructions)
            logger.info(f"Exporting to CSV with these Pandas options: {options}")
            complain_on_unhandled_hints(self.processing_instructions.fail_if_dont_understand,
                                        unhandled_hints, records_format.hints)

            def save_df(df: 'DataFrame', output_filename: str) -> None:
                df.to_csv(path_or_buf=output_filename,
                          index=self.include_index,
                          **options)
                logger.info('CSV file written')
        elif isinstance(records_format, ParquetRecordsFormat):
            # Pyarrow is the only engine we've tested with, and it
            # needed special options, so let's tell Pandas to use it
            pyarrow_args = {
                'coerce_timestamps': None
            }

            def save_df(df: 'DataFrame', output_filename: str) -> None:
                logger.info(f"Writing Parquet file to {output_filename}")
                # Note that this doesn't specify partitioning as of yet -
                # https://app.asana.com/0/1128138765527694/1126615025514407
                df.to_parquet(fname=output_filename,
                              engine='pyarrow',
                              index=self.include_index,
                              **pyarrow_args)
                logger.info('Parquet file written')
        else:
            raise NotImplementedError(f"Teach me how to write to {records_format}")

        return self.serialize_dfs(processing_instructions, records_format, save_df)



from abc import ABCMeta, abstractmethod
from sqlalchemy.schema import CreateColumn
from sqlalchemy.engine.interfaces import Dialect
from sqlalchemy import Column
import json
from collections import OrderedDict
import logging
import re
from typing import Optional, Dict, Union, Any, cast, TYPE_CHECKING
if TYPE_CHECKING:
    from typing_extensions import Literal
    from mypy_extensions import TypedDict
    import pandas

    class FieldRepresentationDict(TypedDict):
        rep_type: str

    class MandatorySqlFieldRepresentationDict(FieldRepresentationDict):
        col_ddl: str

    class SqlFieldRepresentationDict(MandatorySqlFieldRepresentationDict, total=False):
        col_type: str
        col_modifiers: str

    class CorePandasFieldRepresentationDict(FieldRepresentationDict):
        pd_df_dtype: Dict[str, Any]
        pd_df_coltype: Union[Literal['index'], Literal['series']]

    class PandasFieldRepresentationDict(CorePandasFieldRepresentationDict, total=False):
        pd_df_ftype: str


logger = logging.getLogger(__name__)


class RecordsSchemaFieldRepresentation(metaclass=ABCMeta):
    @staticmethod
    def from_data(data: 'FieldRepresentationDict') -> Optional['RecordsSchemaFieldRepresentation']:
        if data['rep_type'].startswith('sql/'):
            sql_data = cast('SqlFieldRepresentationDict', data)
            return RecordsSchemaSqlFieldRepresentation(col_ddl=sql_data['col_ddl'],
                                                       rep_type=sql_data['rep_type'],
                                                       col_type=sql_data.get('col_type'),
                                                       col_modifiers=sql_data.get('col_modifiers'))
        elif data['rep_type'] == 'dataframe/pandas':
            pandas_data = cast('PandasFieldRepresentationDict', data)
            return RecordsSchemaPandasFieldRepresentation(
                pd_df_dtype=pandas_data['pd_df_dtype'],
                pd_df_ftype=pandas_data.get('pd_df_ftype'),
                pd_df_coltype=pandas_data['pd_df_coltype'])
        else:
            logger.warning(f"I don't understand type {data['rep_type']}--"
                           "consider upgrading records-mover")
            return None

    @staticmethod
    def from_series(series: 'pandas.Series') -> 'RecordsSchemaFieldRepresentation':
        import pandas

        dtype_json_str = pandas.io.json.dumps(series.dtype)
        dtype_numpy_rep = json.loads(dtype_json_str)
        ftype = getattr(series, 'ftype', None)
        return RecordsSchemaPandasFieldRepresentation(pd_df_coltype='series',
                                                      pd_df_dtype=dtype_numpy_rep,
                                                      pd_df_ftype=ftype)

    @staticmethod
    def from_index(index: 'pandas.Index') -> 'RecordsSchemaFieldRepresentation':
        dtype_json_str = pandas.io.json.dumps(index.dtype)
        dtype_numpy_rep = json.loads(dtype_json_str)
        return RecordsSchemaPandasFieldRepresentation(pd_df_coltype='index',
                                                      pd_df_dtype=dtype_numpy_rep,
                                                      pd_df_ftype=index.ftype)

    @staticmethod
    def from_sqlalchemy_column(column: Column, dialect: Dialect,
                               rep_type: str) ->\
            'RecordsSchemaFieldRepresentation':
        statement = CreateColumn(column)
        col_ddl_compiler = statement.compile(dialect=dialect)
        full_col_ddl = str(col_ddl_compiler)

        #
        # Parse single string output from SQLAlchemy into what records
        # schema wants.  Sadly there's no universal way to get this
        # out of the SQLAlchemy drivers themselves - example code from
        # the Redshift driver:
        # https://github.com/sqlalchemy-redshift/sqlalchemy-redshift/blob/6885b059c2d23423ae763678a46bbcdd1b3db433/sqlalchemy_redshift/dialect.py#L345
        #
        # These regexps are in order from most specfic to least so
        # that we can match early on the special cases without them
        # being caught by the more general ones.  If this code is not
        # working right for your case, please see the unit test in
        # test_field_representation and enhance it first to cover your
        # newly discovered oddity.
        #

        regexps: Dict[str, str] = OrderedDict()
        # First, cover cases where the types themselves have spaces in them:
        # e.g., DOUBLE PRECISION
        regexps['two_word_type_1'] = r'^([\S]+) (DOUBLE PRECISION)$'
        # e.g,, DOUBLE PRECISION SORTKEY
        regexps['two_word_type_1_with_col_modifiers'] = r'^([\S]+) (DOUBLE PRECISION) (.*)$'
        # e.g., TIMESTAMP WITH TIME ZONE
        regexps['two_word_type_2'] = r'^([\S]+) (TIMESTAMP WITH TIME ZONE)$'
        # e.g., TIMESTAMP WITH TIME ZONE ENCODE lzo
        regexps['two_word_type_2_with_col_modifiers'] = r'^([\S]+) (TIMESTAMP WITH TIME ZONE) (.*)$'
        # e.g., TIMESTAMP WITHOUT TIME ZONE
        regexps['two_word_type_3'] = r'^([\S]+) (TIMESTAMP WITHOUT TIME ZONE)$'
        # e.g., TIMESTAMP WITH TIME ZONE ENCODE lzo
        regexps['two_word_type_3_with_col_modifiers'] =\
            r'^([\S]+) (TIMESTAMP WITHOUT TIME ZONE) (.*)$'
        # e.g., NUMERIC(2, 1)
        regexps['parenthetical_col_type'] = r'^([\S]+) ([\S]+\(.*\))$'
        # e.g., NUMERIC(2, 1) ENCODE lzo
        regexps['parenthetical_col_type_with_col_modifiers'] = r'^([\S]+) ([\S]+\(.*\)) (.*)$'
        # e.g., REAL
        regexps['boring_col_type'] = r'^([\S]+) ([\S]+)$'
        # e.g., REAL ENCODE lzo
        regexps['boring_col_type_with_col_modifiers'] = r'^([\S]+) ([\S]+) (.*)$'

        for regexp in regexps.values():
            match = re.match(regexp, full_col_ddl)
            if match is not None:
                statement_words = list(match.groups())
                break
        else:
            raise SyntaxError(f"Could not understand {full_col_ddl}")
        col_ddl = ' '.join(statement_words[1:])
        col_type = statement_words[1]
        col_modifiers: Optional[str] = ' '.join(statement_words[2:])
        if col_modifiers == '':
            col_modifiers = None
        return RecordsSchemaSqlFieldRepresentation(col_ddl=col_ddl,
                                                   rep_type=rep_type,
                                                   col_type=col_type,
                                                   col_modifiers=col_modifiers)

    @abstractmethod
    def to_data(self) -> 'FieldRepresentationDict':
        raise NotImplementedError


class RecordsSchemaSqlFieldRepresentation(RecordsSchemaFieldRepresentation):
    def __init__(self,
                 col_ddl: str,
                 rep_type: str,
                 col_type: Optional[str],
                 col_modifiers: Optional[str]):
        self.rep_type = rep_type
        self.col_ddl = col_ddl
        self.col_type = col_type
        self.col_modifiers = col_modifiers

    def to_data(self) -> 'SqlFieldRepresentationDict':
        out: 'SqlFieldRepresentationDict' = {
            'rep_type': self.rep_type,
            'col_ddl': self.col_ddl
        }
        if self.col_type is not None:
            out['col_type'] = self.col_type
        if self.col_modifiers is not None:
            out['col_modifiers'] = self.col_modifiers
        return out


class RecordsSchemaPandasFieldRepresentation(RecordsSchemaFieldRepresentation):
    def __init__(self,
                 pd_df_dtype: Dict[str, Any],
                 pd_df_ftype: Optional[str],
                 pd_df_coltype: Union['Literal["series"]', 'Literal["index"]']):
        self.pd_df_dtype = pd_df_dtype
        self.pd_df_ftype = pd_df_ftype
        self.pd_df_coltype = pd_df_coltype

    def to_data(self) -> 'PandasFieldRepresentationDict':
        out: 'PandasFieldRepresentationDict' = {
            'rep_type': 'dataframe/pandas',
            'pd_df_dtype': self.pd_df_dtype,
            'pd_df_coltype': self.pd_df_coltype,
        }
        if self.pd_df_ftype is not None:
            out['pd_df_ftype'] = self.pd_df_ftype
        return out



from typing import NamedTuple, Optional, Mapping


SaveResult = NamedTuple('SaveResults', [('output_urls', Mapping[str, str])])
# export_count may be None if database doesn't support accurate counts
UnloadResult = NamedTuple('UnloadResult', [('export_count', Optional[int])])
# import_count may be None if database doesn't support accurate counts
LoadResult = NamedTuple('LoadResult', [('import_count', Optional[int])])

# move_count and output_urls may be empty depending on the nature of
# the move - e.g., a move to a database doesn't currently map to a
# formal URL, and whole-file based moves do not currently count the
# number of records being moved.
MoveResult = NamedTuple('MoveResult',
                        [('move_count', Optional[int]),
                         ('output_urls', Optional[Mapping[str, str]])])



from subprocess import check_output
from records_mover.db.db_type import canonicalize_db_type, db_protocol
from db_facts.db_facts_types import DBFacts


def lpass_field(name: str, field: str) -> str:
    if field == 'notes':
        field_arg = '--notes'
    elif field == 'username':
        field_arg = '--username'
    elif field == 'password':
        field_arg = '--password'
    elif field == 'url':
        field_arg = '--url'
    else:
        field_arg = '--field=' + field
    raw_output = check_output(['lpass',
                               'show',
                               field_arg,
                               name])
    return raw_output.decode('utf-8').rstrip('\n')


def db_facts_from_lpass(lpass_entry_name: str) -> DBFacts:
    user = lpass_field(lpass_entry_name, 'username')
    password = lpass_field(lpass_entry_name, 'password')
    host = lpass_field(lpass_entry_name, 'Hostname')
    port = int(lpass_field(lpass_entry_name, 'Port'))
    raw_db_type = lpass_field(lpass_entry_name, 'Type')
    db_type = canonicalize_db_type(raw_db_type)
    dbname = lpass_field(lpass_entry_name, 'Database')

    return {'password': password,
            'host': host,
            'user': user,
            'type': db_type,
            'protocol': db_protocol(db_type),
            'port': port,
            'database': dbname}



import inspect
from typing import Optional
from ..types import JsonSchema
from .json_schema_document import JsonSchemaDocument, DefaultValue


class JsonSchemaArrayDocument(JsonSchemaDocument):
    """Represents a JSON Schema array type"""
    def __init__(self,
                 json_type: str,
                 items: JsonSchemaDocument,
                 default: DefaultValue=inspect.Parameter.empty,
                 description: Optional[str]=None) -> None:
        super().__init__(json_type=json_type, default=default, description=description)
        self.items = items

    def to_data(self) -> JsonSchema:
        out = super().to_data()
        out['items'] = self.items.to_data()
        return out



from ..deprecated import warn_deprecated
import pathlib
from contextlib import contextmanager
from sqlalchemy.engine import Engine
from ..records_format import BaseRecordsFormat
from ..schema import RecordsSchema
from ..records_directory import RecordsDirectory
from ...db import DBDriver
from ..processing_instructions import ProcessingInstructions
from ...url.resolver import UrlResolver
from .fileobjs import FileobjsSource  # noqa
from .uninferred_fileobjs import UninferredFileobjsRecordsSource
from .data_url import DataUrlRecordsSource
from .directory import RecordsDirectoryRecordsSource
from .. import RecordsHints, BootstrappingRecordsHints
from .base import (SupportsRecordsDirectory, SupportsMoveToRecordsDirectory,  # noqa
                   SupportsToFileobjsSource, RecordsSource)
from typing import Mapping, IO, Callable, Iterator, Optional, Union, Iterable, TYPE_CHECKING
if TYPE_CHECKING:
    # see the 'gsheets' extras_require option in setup.py - needed for this!
    import google.auth.credentials  # noqa
    from .google_sheets import GoogleSheetsRecordsSource  # noqa ditto
    # with pandas, which an optional addition for clients of this
    # library
    from pandas import DataFrame  # noqa
    from .dataframes import DataframesRecordsSource  # noqa
    from .table import TableRecordsSource  # noqa


class RecordsSources(object):
    """
    These methods produce objects representing the source of a records move.

    Example use:

    .. code-block:: python
       records = session.records
       db_engine = session.get_default_db_engine()
       url = 's3://some-bucket/some-directory/'
       source = records.sources.directory_from_url(url=url)
       target = records.targets.table(schema_name='myschema',
                                      table_name='mytable',
                                      db_engine=db_engine)
       results = records.move(source, target)
    """
    def __init__(self,
                 db_driver: Callable[[Engine], DBDriver],
                 url_resolver: UrlResolver) -> None:
        self.db_driver = db_driver
        self.url_resolver = url_resolver

    @contextmanager
    def dataframe(self,
                  df: 'DataFrame',
                  schema_name: Optional[str]=None,
                  table_name: Optional[str]=None,
                  db_engine: Optional[Engine]=None,
                  processing_instructions: ProcessingInstructions=
                  ProcessingInstructions(),
                  records_schema: Optional[RecordsSchema]=None,
                  include_index: bool=False) -> Iterator['DataframesRecordsSource']:
        """
        :param df: Pandas dataframe to move data from.
        :param schema_name: Obsolete - not used.
        :param table_name: Obsolete - not used.
        :param db_engine: Obsolete - not used.
        :param processing_instructions: Instructions used during creation of the schema SQL.
        :param records_schema: Description of the column names and types of the records.
        :param include_index: If true, the Pandas dataframe index column will be included in
        the move.
        """

        warn_deprecated(schema_name=schema_name, table_name=table_name, db_engine=db_engine)
        from .dataframes import DataframesRecordsSource  # noqa
        yield DataframesRecordsSource(dfs=[df],
                                      records_schema=records_schema,
                                      processing_instructions=processing_instructions,
                                      include_index=include_index)

    @contextmanager
    def dataframes(self,
                   dfs: Iterable['DataFrame'],
                   processing_instructions: ProcessingInstructions=
                   ProcessingInstructions(),
                   records_schema: Optional[RecordsSchema]=None,
                   include_index: bool=False) -> Iterator['DataframesRecordsSource']:
        """

        :param dfs: Iterator of Pandas dataframes to move data from -- all data from
                    these dataframes will be added to the same table.
        :param processing_instructions: Instructions used during creation of the schema SQL.
        :param records_schema: Description of the column names and types of the records.
        :param include_index: If true, the Pandas dataframe index column will be included in
        the move.
        """
        from .dataframes import DataframesRecordsSource  # noqa
        yield DataframesRecordsSource(dfs=dfs,
                                      records_schema=records_schema,
                                      processing_instructions=processing_instructions,
                                      include_index=include_index)

    @contextmanager
    def fileobjs(self,
                 target_names_to_input_fileobjs: Mapping[str, IO[bytes]],
                 records_format: Optional[BaseRecordsFormat]=None,
                 initial_hints: Optional[BootstrappingRecordsHints]=None,
                 records_schema: Optional[RecordsSchema]=None)\
            -> Iterator[Union[UninferredFileobjsRecordsSource, FileobjsSource]]:
        """
        :param target_names_to_input_fileobjs: Filenames mapping to streams of data file.
        :param records_format: Description of the format of the data files.
        :param initial_hints: If records_format is not provided, the format of the file
        will be determined automatically.  If that effort fails, you can help it out by
        providing the 'compression', 'quoting', 'header-row', 'field-delimiter', 'encoding', and
        'escape' hints (or the appropriate subsets) in this dictionary.
        :param records_schema: Description of the column names and types of the records.
        """
        if records_schema is None or records_format is None:
            yield UninferredFileobjsRecordsSource(
                target_names_to_input_fileobjs=target_names_to_input_fileobjs,
                records_format=records_format,
                records_schema=records_schema,
                initial_hints=initial_hints)
        else:
            yield FileobjsSource(
                target_names_to_input_fileobjs=target_names_to_input_fileobjs,
                records_format=records_format,
                records_schema=records_schema)

    @contextmanager
    def data_url(self,
                 input_url: str,
                 records_format: Optional[BaseRecordsFormat]=None,
                 initial_hints: Optional[BootstrappingRecordsHints]=None,
                 records_schema: Optional[RecordsSchema]=None)\
            -> Iterator[DataUrlRecordsSource]:
        """
        :param input_url: Location of the data file.  Must be a URL format understood by the
        records_mover.url library.
        :param records_format: Description of the format of the data files.
        :param initial_hints: If records_format is not provided, the format of the file
        will be determined automatically.  If that effort fails, you can help it out by
        providing the 'compression', 'quoting', 'header-row', 'field-delimiter', 'encoding', and
        'escape' hints (or the appropriate subsets) in this dictionary.
        :param records_schema: Description of the column names and types of the records.
        """
        yield DataUrlRecordsSource(input_url=input_url,
                                   url_resolver=self.url_resolver,
                                   records_format=records_format,
                                   records_schema=records_schema,
                                   initial_hints=initial_hints)

    @contextmanager
    def table(self,
              db_engine: Engine,
              schema_name: str,
              table_name: str) -> Iterator['TableRecordsSource']:
        """
        :param db_engine: Database engine to pull data from.
        :param schema_name: Schema name of a table to get data from.
        :param table_name: Table name of a table to get data from.
        """
        from .table import TableRecordsSource  # noqa
        yield TableRecordsSource(schema_name=schema_name,
                                 table_name=table_name,
                                 url_resolver=self.url_resolver,
                                 driver=self.db_driver(db_engine))

    @contextmanager
    def directory_from_url(self,
                           url: str,
                           hints: RecordsHints={},
                           fail_if_dont_understand: bool=True)\
            -> Iterator[RecordsDirectoryRecordsSource]:
        """
        :param url: Location of the records directory.  Must be a URL format understood by the
        records_mover.url library, and must be a directory URL that ends with a '/'.
        :param hints: Any additional hints that should override the description of the data files
        already in the records directory.
        :param fail_if_dont_understand: If True, and a part of the
        RecordsFormat is not understood while processing, then
        immediately fail and raise an exception.  Otherwise, ignore
        the misunderstood instruction (e.g., ignore the hint, assume
        default variant, etc etc)
        """
        directory_loc = self.url_resolver.directory_url(url)
        directory = RecordsDirectory(records_loc=directory_loc)
        fail = fail_if_dont_understand
        yield RecordsDirectoryRecordsSource(directory=directory,
                                            fail_if_dont_understand=fail,
                                            override_hints=hints,
                                            url_resolver=self.url_resolver)

    @contextmanager
    def local_file(self,
                   filename: str,
                   records_format: Optional[BaseRecordsFormat]=None,
                   initial_hints: Optional[BootstrappingRecordsHints]=None,
                   records_schema: Optional[RecordsSchema]=None)\
            -> Iterator[DataUrlRecordsSource]:
        """
        :param filename: File path (relative or absolute) of the data file to load.
        :param records_format: Description of the format of the data files.
        :param initial_hints: If records_format is not provided, the format of the file
        will be determined automatically.  If that effort fails, you can help it out by
        providing the 'compression', 'quoting', 'header-row', 'field-delimiter', 'encoding', and
        'escape' hints (or the appropriate subsets) in this dictionary.
        :param records_schema: Description of the column names and types of the records.
        """
        url = pathlib.Path(filename).resolve().as_uri()
        with self.data_url(input_url=url,
                           records_format=records_format,
                           records_schema=records_schema,
                           initial_hints=initial_hints) as source:
            yield source

    @contextmanager
    def google_sheet(self,
                     spreadsheet_id: str,
                     sheet_name_or_range: str,
                     google_cloud_creds:
                     'google.auth.credentials.Credentials',
                     out_of_band_column_headers: Optional[Iterable[str]]=None,
                     header_translator: Optional[Callable[[str], str]]=None) ->\
            Iterator['GoogleSheetsRecordsSource']:
        """:param spreadsheet_id: This is the xyz in
        https://docs.google.com/spreadsheets/d/xyz/edit?ts=5be5b383#gid=abc
        :param sheet_name_or_range: This is the label of the particular tab within the Google Sheets
        spreadsheet where the data should go, or a valid Google Sheets-style range formula
        :param google_cloud_creds: This is the dict form of the JSON containing private keys and
                                   client IDs and whatnot for Google Cloud Platform access.
        :param out_of_band_column_headers: If provided, we'll use these
        column names instead of the first row of the spreadsheet.  If
        set, the first row will be treated as data.
        :param header_translator: If provided, header names pulled from
        the sheet will be translated through this function.  Not used
        if out_of_band_column_headers is set.
        """
        from .google_sheets import GoogleSheetsRecordsSource  # noqa
        yield GoogleSheetsRecordsSource(spreadsheet_id=spreadsheet_id,
                                        sheet_name_or_range=sheet_name_or_range,
                                        google_cloud_creds=google_cloud_creds,
                                        out_of_band_column_headers=out_of_band_column_headers,
                                        header_translator=header_translator)



from .sources import (RecordsSource, SupportsMoveToRecordsDirectory,
                      SupportsToFileobjsSource,
                      FileobjsSource, SupportsToDataframesSource)
from .targets.base import (RecordsTarget, SupportsMoveFromRecordsDirectory,
                           SupportsMoveFromTempLocAfterFillingIt,
                           MightSupportMoveFromFileobjsSource,
                           SupportsMoveFromDataframes)
from .sources import base as sources_base
from .targets import base as targets_base
from .processing_instructions import ProcessingInstructions
from .records_format import BaseRecordsFormat
from .results import MoveResult
from typing import Union, ContextManager, TypeVar, cast
import logging

logger = logging.getLogger(__name__)


A = TypeVar('A')


def unpack_context_manager(potential_iterator: Union[A, ContextManager[A]]) -> A:
    if hasattr(potential_iterator, '__enter__'):
        iterator = cast(ContextManager[A], potential_iterator)
        with iterator as unpacked:
            return unpacked
    else:
        a = cast(A, potential_iterator)
        return a


def move(records_source: Union[RecordsSource, ContextManager[RecordsSource]],
         records_target: Union[RecordsTarget, ContextManager[RecordsTarget]],
         processing_instructions: ProcessingInstructions=ProcessingInstructions()) -> MoveResult:
    """Copy records from one location to another.  Applies a sequence of
    possible techniques to do this in an efficient way by looking for
    whether copy-related methods are implementeds on the source and
    target.

    To figure out how to get your new source or target to work well
    with move, follow this function down from the top and implement
    the first method which makes sense for your new source/target.
    See documentation for RecordsSource in sources.py and
    RecordsTarget in targets.py for the semantics of the methods being
    called.
    """
    records_source = unpack_context_manager(records_source)
    records_target = unpack_context_manager(records_target)

    records_source.validate()
    records_target.validate()

    if (isinstance(records_source, sources_base.SupportsRecordsDirectory) and
        isinstance(records_target, SupportsMoveFromRecordsDirectory) and
       records_target.can_load_direct(records_source.records_directory().loc.scheme) and
       records_target.can_move_from_this_format(records_source.records_format)):
        # Tell the destination to load directly from wherever the
        # source is, without needing to make any copies of the data or
        # streaming it through the current box.
        directory = records_source.records_directory()
        logger.info(f"Mover: copying from {records_source} to {records_target} "
                    f"by moving from records directory in {directory}...")
        return records_target.\
            move_from_records_directory(directory=directory,
                                        override_records_format=records_source.records_format,
                                        processing_instructions=processing_instructions)
    elif (isinstance(records_source, FileobjsSource) and
          isinstance(records_target, MightSupportMoveFromFileobjsSource) and
          records_target.can_move_from_fileobjs_source() and
          records_target.can_move_from_this_format(records_source.records_format)):
        logger.info(f"Mover: copying from {records_source} to {records_target} "
                    "by moving directly from stream...")
        # See if we can stream from the source to the destination directly
        return records_target.move_from_fileobjs_source(records_source,
                                                        processing_instructions)
    elif (isinstance(records_source, SupportsMoveToRecordsDirectory) and
          isinstance(records_target, targets_base.SupportsRecordsDirectory) and
          records_source.has_compatible_format(records_target)):
        # if target can accept records and doesn't specify a
        # records_format, or uses the same as the source, we can just
        # dump bytes directly!
        directory = records_target.records_directory()
        logger.info(f"Mover: copying from {records_source} to {records_target} "
                    f"by writing to {directory.scheme} records directory")
        pi = processing_instructions
        records_format = records_source.compatible_format(records_target)
        assert records_format is not None  # we checked compatibility above
        records_target.pre_load_hook()
        out = records_source.\
            move_to_records_directory(records_directory=directory,
                                      records_format=records_format,
                                      processing_instructions=pi)
        records_target.post_load_hook(num_rows_loaded=out.move_count)
        return out
    elif (isinstance(records_source, sources_base.SupportsRecordsDirectory) and
          isinstance(records_target, SupportsMoveFromRecordsDirectory) and
          records_target.can_move_from_this_format(records_source.records_format)):
        directory = records_source.records_directory()
        logger.info(f"Mover: copying from {records_source} to {records_target} "
                    "by loading from {directory.scheme} directory...")
        return records_target.\
            move_from_records_directory(directory=directory,
                                        override_records_format=records_source.records_format,
                                        processing_instructions=processing_instructions)
    elif isinstance(records_source, SupportsToFileobjsSource):
        target_records_format: BaseRecordsFormat = getattr(records_target, "records_format", None)
        logger.info(f"Mover: copying from {records_source} to {records_target} "
                    f"by first writing {records_source} to {target_records_format} "
                    "records format...")
        with records_source.\
                to_fileobjs_source(processing_instructions=processing_instructions,
                                   records_format_if_possible=target_records_format)\
                as fileobjs_source:
            return move(fileobjs_source, records_target, processing_instructions)
    elif (isinstance(records_source, SupportsMoveToRecordsDirectory) and
          isinstance(records_target, SupportsMoveFromTempLocAfterFillingIt) and
          records_source.has_compatible_format(records_target)):
        logger.info(f"Mover: copying from {records_source} to {records_target} "
                    f"by filling in a temporary location...")
        return records_target.move_from_temp_loc_after_filling_it(records_source,
                                                                  processing_instructions)
    elif (isinstance(records_source, SupportsToDataframesSource) and
          isinstance(records_target, SupportsMoveFromDataframes)):
        logger.info(f"Mover: copying from {records_source} to {records_target} "
                    f"by converting to dataframe...")
        with records_source.to_dataframes_source(processing_instructions) as dataframes_source:
            return records_target.\
                move_from_dataframes_source(dfs_source=dataframes_source,
                                            processing_instructions=processing_instructions)
    elif (isinstance(records_source, SupportsToDataframesSource)):
        logger.info(f"Mover: copying from {records_source} to {records_target} "
                    f"by first converting to dataframe...")
        with records_source.to_dataframes_source(processing_instructions) as dataframes_source:
            return move(dataframes_source, records_target, processing_instructions)
    else:
        raise NotImplementedError(f"Please teach me how to move records between "
                                  f"{records_source} and {records_target}")



import boto3
from db_facts import db
from .lpass import lpass_field
import json
from typing import Iterable
from .base_creds import BaseCreds
from typing import TYPE_CHECKING
from db_facts.db_facts_types import DBFacts
if TYPE_CHECKING:
    # see the 'gsheets' extras_require option in setup.py - needed for this!
    import google.auth.credentials  # noqa


class CredsViaLastPass(BaseCreds):
    def _gcp_creds(self, gcp_creds_name: str,
                   scopes: Iterable[str]) -> 'google.auth.credentials.Credentials':
        import google.oauth2.service_account
        notes_json = lpass_field(gcp_creds_name, 'notes')
        cred_details = json.loads(notes_json)

        return google.oauth2.service_account.Credentials.\
            from_service_account_info(cred_details, scopes=scopes)

    def db_facts(self, db_creds_name: str) -> DBFacts:
        return db(db_creds_name.split('-'))

    def boto3_session(self, aws_creds_name: str) -> boto3.session.Session:
        if aws_creds_name is None:
            return boto3.session.Session()
        else:
            raise NotImplementedError



import csv
from ...utils import quiet_remove
from ..hints import cant_handle_hint
from ..processing_instructions import ProcessingInstructions
from ..records_format import DelimitedRecordsFormat
import logging
from typing import Set, Dict, Any


logger = logging.getLogger(__name__)


def pandas_read_csv_options(records_format: DelimitedRecordsFormat,
                            unhandled_hints: Set[str],
                            processing_instructions: ProcessingInstructions) -> Dict[str, Any]:
    ...
    # https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.read_csv.html#pandas.read_csv

    hints = records_format.hints

    fail_if_cant_handle_hint = processing_instructions.fail_if_cant_handle_hint

    pandas_options: Dict[str, object] = {}

    #
    # filepath_or_buffer : str, path object, or file-like object
    #
    # Any valid string path is acceptable. The string could be a
    # URL. Valid URL schemes include http, ftp, s3, and file. For file
    # URLs, a host is expected. A local file could be:
    # file://localhost/path/to/table.csv.
    #
    # If you want to pass in a path object, pandas accepts either
    # pathlib.Path or py._path.local.LocalPath.
    #
    # By file-like object, we refer to objects with a read() method,
    # such as a file handler (e.g. via builtin open function) or StringIO.
    #

    #
    # sep : str, default ‘,’
    #
    # (added by caller to this function)
    #
    # Delimiter to use. If sep is None, the C engine cannot
    # automatically detect the separator, but the Python parsing
    # engine can, meaning the latter will be used and automatically
    # detect the separator by Python’s builtin sniffer tool,
    # csv.Sniffer. In addition, separators longer than 1 character and
    # different from '\s+' will be interpreted as regular expressions
    # and will also force the use of the Python parsing engine. Note
    # that regex delimiters are prone to ignoring quoted data. Regex
    # example: '\r\t'.
    #

    #
    # delimiter : str, default None
    #
    # Alias for sep.
    #
    pandas_options['delimiter'] = hints['field-delimiter']
    quiet_remove(unhandled_hints, 'field-delimiter')

    #
    # header : int, list of int, default ‘infer’
    #
    # Row number(s) to use as the column names, and the start of the
    # data. Default behavior is to infer the column names: if no names
    # are passed the behavior is identical to header=0 and column
    # names are inferred from the first line of the file, if column
    # names are passed explicitly then the behavior is identical to
    # header=None. Explicitly pass header=0 to be able to replace
    # existing names. The header can be a list of integers that
    # specify row locations for a multi-index on the columns
    # e.g. [0,1,3]. Intervening rows that are not specified will be
    # skipped (e.g. 2 in this example is skipped). Note that this
    # parameter ignores commented lines and empty lines if
    # skip_blank_lines=True, so header=0 denotes the first line of
    # data rather than the first line of the file.
    #
    if hints['header-row']:
        pandas_options['header'] = 0
    else:
        pandas_options['header'] = None
    quiet_remove(unhandled_hints, 'header-row')

    #
    # names : array-like, optional
    #
    # List of column names to use. If file contains no header row,
    # then you should explicitly pass header=None. Duplicates in this
    # list will cause a UserWarning to be issued.
    #

    # (this may be useful when we support reading CSVs into dataframes
    # using an existing records schema that specifies column
    # names--but we don't have that now)

    #
    # index_col : int, sequence or bool, optional
    #
    # Column to use as the row labels of the DataFrame. If a sequence
    # is given, a MultiIndex is used. If you have a malformed file
    # with delimiters at the end of each line, you might consider
    # index_col=False to force pandas to not use the first column as
    # the index (row names).
    #

    # (we don't use dataframe indexes currently; any primary key
    # information would be transferred and handled via records schema)

    #
    # usecols : list-like or callable, optional
    #
    # Return a subset of the columns. If list-like, all elements must
    # either be positional (i.e. integer indices into the document
    # columns) or strings that correspond to column names provided
    # either by the user in names or inferred from the document header
    # row(s). For example, a valid list-like usecols parameter would
    # be [0, 1, 2] or ['foo', 'bar', 'baz']. Element order is ignored,
    # so usecols=[0, 1] is the same as [1, 0]. To instantiate a
    # DataFrame from data with element order preserved use
    # pd.read_csv(data, usecols=['foo', 'bar'])[['foo', 'bar']] for
    # columns in ['foo', 'bar'] order or pd.read_csv(data,
    # usecols=['foo', 'bar'])[['bar', 'foo']] for ['bar', 'foo']
    # order.
    #
    # If callable, the callable function will be evaluated against the
    # column names, returning names where the callable function
    # evaluates to True. An example of a valid callable argument would
    # be lambda x: x.upper() in ['AAA', 'BBB', 'DDD']. Using this
    # parameter results in much faster parsing time and lower memory
    # usage.
    #

    # (no current records mover facility to downselect columns during move)

    #
    # squeeze : bool, default False
    #
    # If the parsed data only contains one column then return a Series.
    #
    # (better to keep a standard format, no matter how many columsn)
    #

    #
    # prefix : str, optional
    #
    # Prefix to add to column numbers when no header, e.g. ‘X’ for X0, X1,
    #

    #
    # Not sure this actually does anything - when loading a CSV format
    # file with an empty final column name - e.g.,
    # tests/integration/resources/delimited-csv-with-header.csv - the
    # column still comes out as 'unnamed: 11'ead as 'untitled_11'.
    #
    # Leaving this in case a future version of Pandas behaves
    # better.
    #
    pandas_options['prefix'] = 'untitled_'

    #
    # mangle_dupe_cols : bool, default True
    #
    # Duplicate columns will be specified as ‘X’, ‘X.1’, …’X.N’,
    # rather than ‘X’…’X’. Passing in False will cause data to be
    # overwritten if there are duplicate names in the columns.
    #

    # (default is smart to avoid confusing SQL databases, which need
    # unique common names)

    #
    # dtype : Type name or dict of column -> type, optional
    #
    # Data type for data or columns. E.g. {‘a’: np.float64, ‘b’:
    # np.int32, ‘c’: ‘Int64’}
    #
    # Use str or object together with suitable na_values settings to
    # preserve and not interpret dtype. If converters are specified,
    # they will be applied INSTEAD of dtype conversion.
    #

    # (not supported now, but this would be great for using
    # sniffed/pre-supplied records schema JSON)

    #
    # engine : {‘c’, ‘python’}, optional
    #
    # Parser engine to use. The C engine is faster while the python
    # engine is currently more feature-complete.
    #

    non_standard_record_terminator = hints['record-terminator'] not in ['\n', '\r\n', '\r']

    if non_standard_record_terminator:
        # the 'lineterminator' option below is only valid for c parser
        pandas_options['engine'] = 'c'
    else:
        pandas_options['engine'] = 'python'

    # (stick with the default until we have reason to change or
    # dynamically adapt)

    #
    # converters : dict, optional
    #
    # Dict of functions for converting values in certain columns. Keys
    # can either be integers or column labels.
    #

    # (maybe a way we can support a richer set of records hints in the
    # future?)

    #
    # true_values : list, optional
    #
    # Values to consider as True.
    #

    # (not currently specified by records hints - maybe it should be)

    #
    # false_values : list, optional
    #
    # Values to consider as False.
    #

    # (not currently specified by records hints - maybe it should be)

    #
    # skipinitialspace : bool, default False
    #
    # Skip spaces after delimiter.
    #

    # (not currently specified by records hints - maybe it should be)

    #
    # skiprows : list-like, int or callable, optional
    #
    # Line numbers to skip (0-indexed) or number of lines to skip
    # (int) at the start of the file.
    #
    # If callable, the callable function will be evaluated against the
    # row indices, returning True if the row should be skipped and
    # False otherwise. An example of a valid callable argument would
    # be lambda x: x in [0, 2].
    #

    # (this may be useful when chunking through a file to avoid
    # blowing up memory)

    #
    # skipfooter : int, default 0
    #
    # Number of lines at bottom of file to skip (Unsupported with
    # engine=’c’).
    #

    # (this may be useful when chunking through a file to avoid
    # blowing up memory)

    #
    # nrows : int, optional
    #
    # Number of rows of file to read. Useful for reading pieces of large files.
    #

    # (this may be useful when chunking through a file to avoid
    # blowing up memory)

    #
    # na_values : scalar, str, list-like, or dict, optional
    #
    # Additional strings to recognize as NA/NaN. If dict passed,
    # specific per-column NA values. By default the following values
    # are interpreted as NaN: ‘’, ‘#N/A’, ‘#N/A N/A’, ‘#NA’,
    # ‘-1.#IND’, ‘-1.#QNAN’, ‘-NaN’, ‘-nan’, ‘1.#IND’, ‘1.#QNAN’,
    # ‘N/A’, ‘NA’, ‘NULL’, ‘NaN’, ‘n/a’, ‘nan’, ‘null’.
    #

    # (not fully specified in spec yet, though some work was started
    # at one point)

    #
    # keep_default_na : bool, default True
    #
    # Whether or not to include the default NaN values when parsing
    # the data. Depending on whether na_values is passed in, the
    # behavior is as follows:
    #
    # If keep_default_na is True, and na_values are specified,
    # na_values is appended to the default NaN values used for
    # parsing.
    #
    # If keep_default_na is True, and na_values are not specified,
    # only the default NaN values are used for parsing.
    #
    # If keep_default_na is False, and na_values are specified, only
    # the NaN values specified na_values are used for parsing.
    #
    # If keep_default_na is False, and na_values are not specified, no
    # strings will be parsed as NaN.
    #
    # Note that if na_filter is passed in as False, the
    # keep_default_na and na_values parameters will be ignored.
    #

    # (not fully specified in spec yet, though some work was started
    # at one point)

    #
    # na_filter : bool, default True
    #
    # Detect missing value markers (empty strings and the value of
    # na_values). In data without any NAs, passing na_filter=False can
    # improve the performance of reading a large file.
    #

    # (not fully specified in spec yet, though some work was started
    # at one point)

    #
    # verbose : bool, default False
    #
    # Indicate number of NA values placed in non-numeric columns.
    #

    # (might be useful if we provided some easy verbosity knobs when
    # using records mover)

    #
    # skip_blank_lines : bool, default True
    #
    # If True, skip over blank lines rather than interpreting as NaN
    # values.
    #

    # (maybe someday we'll have a relevant hint, but for now, use
    # whatever pandas' default behavior is)

    #
    # parse_dates : bool or list of int or names or list of lists or
    # dict, default False
    #
    # The behavior is as follows:
    #
    # boolean. If True -> try parsing the index.
    #
    # list of int or names. e.g. If [1, 2, 3] -> try parsing columns
    # 1, 2, 3 each as a separate date column.
    #
    # list of lists. e.g. If [[1, 3]] -> combine columns 1 and 3 and
    # parse as a single date column.
    #
    # dict, e.g. {‘foo’ : [1, 3]} -> parse columns 1, 3 as date and
    # call result ‘foo’
    #
    # If a column or index cannot be represented as an array of
    # datetimes, say because of an unparseable value or a mixture of
    # timezones, the column or index will be returned unaltered as an
    # object data type. For non-standard datetime parsing, use
    # pd.to_datetime after pd.read_csv. To parse an index or column
    # with a mixture of timezones, specify date_parser to be a
    # partially-applied pandas.to_datetime() with utc=True. See
    # Parsing a CSV with mixed Timezones for more.
    #
    # Note: A fast-path exists for iso8601-formatted dates.
    #

    # (we don't yet pass in a records schema which would provide
    # ability to know in advance which columns are datetimes--sounds
    # like it may be very helpful to do so!)

    quiet_remove(unhandled_hints, 'dateformat')
    quiet_remove(unhandled_hints, 'timeonlyformat')
    quiet_remove(unhandled_hints, 'datetimeformat')
    quiet_remove(unhandled_hints, 'datetimeformattz')

    #
    # infer_datetime_format : bool, default False
    #
    # If True and parse_dates is enabled, pandas will attempt to infer
    # the format of the datetime strings in the columns, and if it can
    # be inferred, switch to a faster method of parsing them. In some
    # cases this can increase the parsing speed by 5-10x.
    #

    # (won't be used since we're not yet able to pass in parse_dates)

    #
    # keep_date_col : bool, default False
    #
    # If True and parse_dates specifies combining multiple columns
    # then keep the original columns.
    #

    # (N/A as we don't specify combining multiple columns in
    # parse_dates)

    #
    # date_parser : function, optional
    #
    # Function to use for converting a sequence of string columns to
    # an array of datetime instances. The default uses
    # dateutil.parser.parser to do the conversion. Pandas will try to
    # call date_parser in three different ways, advancing to the next
    # if an exception occurs: 1) Pass one or more arrays (as defined
    # by parse_dates) as arguments; 2) concatenate (row-wise) the
    # string values from the columns defined by parse_dates into a
    # single array and pass that; and 3) call date_parser once for
    # each row using one or more strings (corresponding to the columns
    # defined by parse_dates) as arguments.
    #

    # (N/A as we don't pass anything as parse_dates)

    #
    # dayfirst : bool, default False
    #
    # DD/MM format dates, international and European format.
    #

    # (N/A as we don't pass anything as parse_dates)

    #
    # iterator : bool, default False
    #
    # Return TextFileReader object for iteration or getting chunks
    # with get_chunk()
    #

    # (may be very useful for reading large files - see
    # https://stackoverflow.com/questions/39386458/how-to-read-data-in-python-dataframe-without-concatenating)

    #
    # chunksize : int, optional
    #
    # Return TextFileReader object for iteration. See the IO Tools
    # docs for more information on iterator and chunksize.
    #

    # (may be very useful for reading large files - see
    # https://stackoverflow.com/questions/39386458/how-to-read-data-in-python-dataframe-without-concatenating)

    #
    # compression : {‘infer’, ‘gzip’, ‘bz2’, ‘zip’, ‘xz’, None},
    # default ‘infer’
    #
    # For on-the-fly decompression of on-disk data. If ‘infer’ and
    # filepath_or_buffer is path-like, then detect compression from
    # the following extensions: ‘.gz’, ‘.bz2’, ‘.zip’, or ‘.xz’
    # (otherwise no decompression). If using ‘zip’, the ZIP file must
    # contain only one data file to be read in. Set to None for no
    # decompression.
    #
    # New in version 0.18.1: support for ‘zip’ and ‘xz’ compression.
    #
    if hints['compression'] is None:
        # hints['compression']=None will output an uncompressed csv,
        # which is the pandas default.
        pandas_options['compression'] = None
    elif hints['compression'] == 'GZIP':
        pandas_options['compression'] = 'gzip'
    elif hints['compression'] == 'BZIP':
        pandas_options['compression'] = 'bz2'
    else:
        cant_handle_hint(fail_if_cant_handle_hint, 'compression', hints)
    quiet_remove(unhandled_hints, 'compression')

    #
    # thousands : str, optional
    #
    # Thousands separator.
    #
    # decimal : str, default ‘.’
    #

    # (not currently specified by records hints - maybe it should be)

    #
    # Character to recognize as decimal point (e.g. use ‘,’ for
    # European data).
    #

    # (not currently specified by records hints - maybe it should be)

    #
    # lineterminator : str (length 1), optional
    #
    # Character to break file into lines. Only valid with C parser.
    #
    if non_standard_record_terminator:
        pandas_options['lineterminator'] = hints['record-terminator']
    quiet_remove(unhandled_hints, 'record-terminator')

    #
    # quotechar : str (length 1), optional
    #
    # The character used to denote the start and end of a quoted
    # item. Quoted items can include the delimiter and it will be
    # ignored.
    #
    pandas_options['quotechar'] = hints['quotechar']
    quiet_remove(unhandled_hints, 'quotechar')

    #
    # quoting : int or csv.QUOTE_* instance, default 0
    #
    # Control field quoting behavior per csv.QUOTE_* constants. Use
    # one of QUOTE_MINIMAL (0), QUOTE_ALL (1), QUOTE_NONNUMERIC (2) or
    # QUOTE_NONE (3).
    #

    if hints['quoting'] is None:
        pandas_options['quoting'] = csv.QUOTE_NONE
    elif hints['quoting'] == 'all':
        pandas_options['quoting'] = csv.QUOTE_ALL
    elif hints['quoting'] == 'minimal':
        pandas_options['quoting'] = csv.QUOTE_MINIMAL
    elif hints['quoting'] == 'nonnumeric':
        pandas_options['quoting'] = csv.QUOTE_NONNUMERIC
    else:
        cant_handle_hint(fail_if_cant_handle_hint, 'quoting', hints)
    quiet_remove(unhandled_hints, 'quoting')

    #
    # doublequote : bool, default True
    #
    # When quotechar is specified and quoting is not QUOTE_NONE,
    # indicate whether or not to interpret two consecutive quotechar
    # elements INSIDE a field as a single quotechar element.
    #
    pandas_options['doublequote'] = hints['doublequote']
    quiet_remove(unhandled_hints, 'doublequote')

    #
    # escapechar : str (length 1), optional
    #
    # One-character string used to escape other characters.
    #
    if hints['escape'] is None:
        pass
    else:
        pandas_options['escapechar'] = hints['escape']
    quiet_remove(unhandled_hints, 'escape')

    #
    # comment : str, optional
    #
    # Indicates remainder of line should not be parsed. If found at
    # the beginning of a line, the line will be ignored
    # altogether. This parameter must be a single character. Like
    # empty lines (as long as skip_blank_lines=True), fully commented
    # lines are ignored by the parameter header but not by
    # skiprows. For example, if comment='#', parsing
    # #empty\na,b,c\n1,2,3 with header=0 will result in ‘a,b,c’ being
    # treated as the header.
    #

    # (not currently specified by records hints - maybe it should be)

    #
    # encoding : str, optional
    #
    # Encoding to use for UTF when reading/writing (ex. ‘utf-8’). List
    # of Python standard encodings .
    #

    # should only be specified when reading from a filepath
    #
    if hints['compression'] is not None:
        pandas_options['encoding'] = hints['encoding']
    quiet_remove(unhandled_hints, 'encoding')

    #
    # dialect : str or csv.Dialect, optional
    #
    # If provided, this parameter will override values (default or
    # not) for the following parameters: delimiter, doublequote,
    # escapechar, skipinitialspace, quotechar, and quoting. If it is
    # necessary to override values, a ParserWarning will be
    # issued. See csv.Dialect documentation for more details.
    #

    # (we specify the other arguments directly except for
    # skipinitialspace, which isn't currently supported by hints)

    #
    # tupleize_cols : bool, default False
    #
    # Leave a list of tuples on columns as is (default is to convert
    # to a MultiIndex on the columns).
    #
    # Deprecated since version 0.21.0: This argument will be removed
    # and will always convert to MultiIndex
    #

    # (deprecated, so not supplying)

    #
    # error_bad_lines : bool, default True
    #
    # Lines with too many fields (e.g. a csv line with too many
    # commas) will by default cause an exception to be raised, and no
    # DataFrame will be returned. If False, then these “bad lines”
    # will dropped from the DataFrame that is returned.
    #

    pandas_options['error_bad_lines'] = processing_instructions.fail_if_row_invalid

    #
    # warn_bad_lines : bool, default True
    #
    # If error_bad_lines is False, and warn_bad_lines is True, a
    # warning for each “bad line” will be output.
    #

    pandas_options['warn_bad_lines'] = True

    #
    # delim_whitespace : bool, default False
    #
    # Specifies whether or not whitespace (e.g. ' ' or ' ') will be
    # used as the sep. Equivalent to setting sep='\s+'. If this option
    # is set to True, nothing should be passed in for the delimiter
    # parameter.
    #
    # New in version 0.18.1: support for the Python parser.
    #

    # (not currently expressible in hints)

    #
    # low_memory : bool, default True
    #
    # Internally process the file in chunks, resulting in lower memory
    # use while parsing, but possibly mixed type inference. To ensure
    # no mixed types either set False, or specify the type with the
    # dtype parameter. Note that the entire file is read into a single
    # DataFrame regardless, use the chunksize or iterator parameter to
    # return the data in chunks. (Only valid with C parser).
    #

    # (We'll see if we see issues without changing the default on this
    # one)

    #
    # memory_map : bool, default False
    #
    # If a filepath is provided for filepath_or_buffer, map the file
    # object directly onto memory and access the data directly from
    # there. Using this option can improve performance because there
    # is no longer any I/O overhead.
    #

    # (I suspect there's a reason this isn't the default, but it may
    # be a knob to play with if we need more performance...)

    #
    # float_precision : str, optional
    #
    # Specifies which converter the C engine should use for
    # floating-point values. The options are None for the ordinary
    # converter, high for the high-precision converter, and round_trip
    # for the round-trip converter.
    #

    # (something to look for if we need more performance...)

    return pandas_options



import sqlalchemy
import logging
from typing import Union, Optional
from sqlalchemy.sql import text

logger = logging.getLogger(__name__)


select_load_errors = """
        select d.query, substring(d.filename,14,20),
        d.line_number as line,
        substring(d.value,1,16) as value,
        substring(le.err_reason,1,100) as err_reason
        from stl_loaderror_detail d, stl_load_errors le
        where d.query = le.query
        and d.query = :last_copy_id;
        """


def schema_sql_from_admin_views(schema: str,
                                table: str,
                                db: Union[sqlalchemy.engine.Engine, sqlalchemy.engine.Connection])\
        -> Optional[str]:
    # The default behavior in current sqlalchemy driver seems to
    # map "DOUBLE PRECISION" to "DOUBLE_PRECISION", so lean back
    # on admin views for now:
    sql = text("""
SELECT ddl
FROM admin.v_generate_tbl_ddl
WHERE schemaname = :schema_name
AND tablename = :table_name
""")
    out: Optional[str]
    try:
        result = db.execute(sql, schema_name=schema, table_name=table).fetchall()
        if len(result) == 0:
            out = None
        else:
            # First line is a commented-out drop table, don't need that.
            out = "\n".join(map(lambda r: r['ddl'], result[1:]))
    except sqlalchemy.exc.ProgrammingError:
        logger.debug("Error while generating SQL", exc_info=True)
        out = None

    if out is None:
        logger.warning("To be able to save SQL to a records directory, please install "
                       "and grant access to 'admin.v_generate_tbl_ddl' from "
                       "https://github.com/awslabs/amazon-redshift-utils/"
                       "blob/master/src/AdminViews/v_generate_tbl_ddl.sql")
    return out



from .database import db_engine, db_facts_from_env
from sqlalchemy.engine import Engine
import boto3
from .records.records import Records
from .db import db_driver, DBDriver
from .url.base import BaseFileUrl, BaseDirectoryUrl
import sqlalchemy
from typing import Union, Optional
from .creds.base_creds import BaseCreds
from .db.connect import engine_from_db_facts
from .url.resolver import UrlResolver
from db_facts.db_facts_types import DBFacts
from enum import Enum
from records_mover.creds.creds_via_lastpass import CredsViaLastPass
from records_mover.creds.creds_via_airflow import CredsViaAirflow
from records_mover.creds.creds_via_env import CredsViaEnv
import subprocess
import os
import logging


logger = logging.getLogger(__name__)


def _infer_session_type() -> str:
    if 'RECORDS_MOVER_SESSION_TYPE' in os.environ:
        return os.environ['RECORDS_MOVER_SESSION_TYPE']

    if 'AIRFLOW__CORE__EXECUTOR' in os.environ:
        # Guess based on an env variable sometimes set by Airflow
        return 'airflow'

    return 'cli'


def _infer_scratch_s3_url(session_type: str) -> Optional[str]:
    scratch_s3_url: Optional[str]

    if "SCRATCH_S3_URL" in os.environ:
        return os.environ["SCRATCH_S3_URL"]

    if session_type == 'cli':
        try:
            #
            # https://app.asana.com/0/1128138765527694/1163219515343393
            #
            # This method of configuration needs to be replaced with
            # something more conventional and documented.
            #
            return subprocess.check_output("scratch-s3-url").decode('ASCII').rstrip()
        except FileNotFoundError:
            pass
    return None


def _infer_default_aws_creds_name(session_type: str) -> Optional[str]:
    if session_type == 'airflow':
        return 'aws_default'
    return None


def _infer_creds(session_type: str) -> BaseCreds:
    if session_type == 'airflow':
        return CredsViaAirflow()
    elif session_type == 'cli':
        #
        # https://app.asana.com/0/1128138765527694/1163219515343393
        #
        # Most people don't use LastPass; other secrets managements
        # should be supported and configurable at the system- and
        # user- level.
        #
        return CredsViaLastPass()
    elif session_type == 'itest':
        return CredsViaEnv()
    elif session_type == 'env':
        return CredsViaEnv()
    elif session_type is not None:
        raise ValueError("Valid job context types: cli, airflow, docker-itest, env - "
                         "consider upgrading records-mover if you're looking for "
                         f"{session_type}.")


class PleaseInfer(Enum):
    # This is a mypy-friendly way of doing a singleton object:
    #
    # https://github.com/python/typing/issues/236
    token = 1


class Session():
    def __init__(self,
                 default_db_creds_name: Optional[str] = None,
                 default_aws_creds_name: Union[None, str, PleaseInfer] = PleaseInfer.token,
                 session_type: Union[str, PleaseInfer] = PleaseInfer.token,
                 scratch_s3_url: Union[None, str, PleaseInfer] = PleaseInfer.token,
                 creds: Union[BaseCreds, PleaseInfer] = PleaseInfer.token) -> None:
        if session_type is PleaseInfer.token:
            session_type = _infer_session_type()

        if creds is PleaseInfer.token:
            creds = _infer_creds(session_type)

        if scratch_s3_url is PleaseInfer.token:
            scratch_s3_url = _infer_scratch_s3_url(session_type)

        if default_aws_creds_name is PleaseInfer.token:
            default_aws_creds_name = _infer_default_aws_creds_name(session_type)

        self._default_db_creds_name = default_db_creds_name
        self._default_aws_creds_name = default_aws_creds_name
        self._scratch_s3_url = scratch_s3_url
        self.creds = creds
        self.url_resolver = UrlResolver(boto3_session=self._boto3_session())

    def get_default_db_engine(self) -> Engine:
        if self._default_db_creds_name is None:
            return db_engine(self)
        else:
            return self.get_db_engine(self._default_db_creds_name)

    def get_default_db_facts(self) -> DBFacts:
        if self._default_db_creds_name is None:
            return db_facts_from_env()
        else:
            return self.creds.db_facts(self._default_db_creds_name)

    def get_db_engine(self,
                      db_creds_name: str,
                      creds_provider: Optional[BaseCreds] = None) -> Engine:
        if creds_provider is None:
            creds_provider = self.creds
        db_facts = creds_provider.db_facts(db_creds_name)
        return engine_from_db_facts(db_facts)

    def db_driver(self, db: Union[sqlalchemy.engine.Engine,
                                  sqlalchemy.engine.Connection]) -> DBDriver:
        s3_temp_base_loc = None
        if self._scratch_s3_url is not None:
            s3_temp_base_loc = self.directory_url(self._scratch_s3_url)
        return db_driver(db=db,
                         s3_temp_base_loc=s3_temp_base_loc,
                         url_resolver=self.url_resolver)

    def file_url(self, url: str) -> BaseFileUrl:
        return self.url_resolver.file_url(url)

    def directory_url(self, url: str) -> BaseDirectoryUrl:
        return self.url_resolver.directory_url(url)

    def _boto3_session(self) -> boto3.session.Session:
        if self._default_aws_creds_name is None:
            return boto3.session.Session()
        else:
            return self.creds.boto3_session(self._default_aws_creds_name)

    @property
    def records(self) -> Records:
        return Records(db_driver=self.db_driver,
                       url_resolver=self.url_resolver)



from sqlalchemy import MetaData
from ..records.unload_plan import RecordsUnloadPlan
from ..records.records_format import BaseRecordsFormat
from ..records.records_directory import RecordsDirectory
from typing import Union, List
import sqlalchemy


class Unloader:
    def __init__(self,
                 db: Union[sqlalchemy.engine.Connection, sqlalchemy.engine.Engine]) -> None:
        self.db = db
        self.meta = MetaData()

    def unload(self,
               schema: str,
               table: str,
               unload_plan: RecordsUnloadPlan,
               directory: RecordsDirectory) -> int:
        raise NotImplementedError(f"Teach me how to bulk export from {self.db.engine.name}")

    def known_supported_records_formats_for_unload(self) -> List[BaseRecordsFormat]:
        return []

    def can_unload_this_format(self, target_records_format: BaseRecordsFormat) -> bool:
        return False



from typing import Union, List, IO, Tuple, Optional
from ...records.hints import complain_on_unhandled_hints
import pprint
import sqlalchemy
from ...records.load_plan import RecordsLoadPlan
from ...records.records_format import (
    BaseRecordsFormat, DelimitedRecordsFormat, ParquetRecordsFormat
)
from ...records.records_directory import RecordsDirectory
from ...records.processing_instructions import ProcessingInstructions
from ...url.resolver import UrlResolver
from google.cloud.bigquery.dbapi.connection import Connection
from google.cloud.bigquery.client import Client
from .load_job_config_options import load_job_config
import logging

logger = logging.getLogger(__name__)


class BigQueryLoader:
    def __init__(self,
                 db: Union[sqlalchemy.engine.Connection, sqlalchemy.engine.Engine],
                 url_resolver: UrlResolver) -> None:
        self.db = db
        self.url_resolver = url_resolver

    def _parse_bigquery_schema_name(self, schema: str) -> Tuple[Optional[str], str]:
        # https://github.com/mxmzdlv/pybigquery/blob/master/pybigquery/sqlalchemy_bigquery.py#L320
        dataset = None
        project = None

        schema_split = schema.split('.')
        if len(schema_split) == 1:
            dataset, = schema_split
        elif len(schema_split) == 2:
            project, dataset = schema_split
        else:
            raise ValueError(f"Could not understand schema name {schema}")

        return (project, dataset)

    def load_from_fileobj(self, schema: str, table: str,
                          load_plan: RecordsLoadPlan, fileobj: IO[bytes]) -> int:
        # https://googleapis.github.io/google-cloud-python/latest/bigquery/usage/tables.html#creating-a-table
        connection: Connection =\
            self.db.engine.raw_connection().connection
        # https://google-cloud.readthedocs.io/en/latest/bigquery/generated/google.cloud.bigquery.client.Client.html
        client: Client = connection._client
        project_id, dataset_id = self._parse_bigquery_schema_name(schema)
        dataset_ref = client.dataset(dataset_id, project_id)
        table_ref = dataset_ref.table(table)
        # https://googleapis.github.io/google-cloud-python/latest/bigquery/generated/google.cloud.bigquery.job.LoadJobConfig.html

        target_records_format = load_plan.records_format
        processing_instructions = load_plan.processing_instructions
        unhandled_hints = set()
        if isinstance(target_records_format, DelimitedRecordsFormat):
            unhandled_hints = set(target_records_format.hints.keys())
        job_config = load_job_config(unhandled_hints, load_plan)
        if isinstance(target_records_format, DelimitedRecordsFormat):
            complain_on_unhandled_hints(processing_instructions.fail_if_dont_understand,
                                        unhandled_hints, target_records_format.hints)
        logger.info(f"Using BigQuery load options: {job_config.to_api_repr()}")
        job = client.load_table_from_file(fileobj,
                                          table_ref,
                                          # Must match the destination dataset location.
                                          location="US",
                                          job_config=job_config)

        try:
            job.result()  # Waits for table load to complete.
        except Exception:
            logger.error(f"BigQuery load errors:\n\n{pprint.pformat(job.errors)}\n")
            raise

        logger.info(f"Loaded {job.output_rows} rows into {dataset_id}:{table}")
        assert job.output_rows is not None  # should be populated after job result is obtained
        return job.output_rows

    def load(self,
             schema: str,
             table: str,
             load_plan: RecordsLoadPlan,
             directory: RecordsDirectory) -> int:
        all_urls = directory.manifest_entry_urls()

        total_rows = 0

        for url in all_urls:
            loc = self.url_resolver.file_url(url)
            with loc.open() as f:
                # We set WriteDisposition.WRITE_APPEND in
                # load_options_config.py and let the records Prep
                # class decide what to do about the existing table, so
                # it's safe to call this multiple times and append until
                # done:
                logger.info(f"Loading {url} into {schema}.{table}")
                total_rows += self.load_from_fileobj(schema, table, load_plan, f)
        return total_rows

    def can_load_this_format(self, source_records_format: BaseRecordsFormat) -> bool:
        try:
            processing_instructions = ProcessingInstructions()
            load_plan = RecordsLoadPlan(records_format=source_records_format,
                                        processing_instructions=processing_instructions)
            if not isinstance(load_plan.records_format, DelimitedRecordsFormat):
                return False
            unhandled_hints = set(load_plan.records_format.hints.keys())
            load_job_config(unhandled_hints, load_plan)
            complain_on_unhandled_hints(processing_instructions.fail_if_dont_understand,
                                        unhandled_hints, load_plan.records_format.hints)
            return True
        except NotImplementedError:
            return False

    def known_supported_records_formats_for_load(self) -> List[BaseRecordsFormat]:
        return [DelimitedRecordsFormat(variant='bigquery'), ParquetRecordsFormat()]



from urllib.parse import urlparse
from .base import BaseDirectoryUrl, BaseFileUrl
from typing import Dict, Type, Union, Callable, Any


CreatesUrls = Callable[..., Union[BaseFileUrl, BaseDirectoryUrl]]

# These are added as modules are evaluated to avoid circular
# dependencies
directory_url_ctors: Dict[str, Union[Type[BaseDirectoryUrl], CreatesUrls]] = {}
file_url_ctors: Dict[str, Union[Type[BaseFileUrl], CreatesUrls]] = {}


def init_urls() -> None:
    from .s3.s3_url import S3Url
    from .filesystem import FilesystemDirectoryUrl, FilesystemFileUrl
    from .http import HttpFileUrl
    if len(directory_url_ctors) == 0:
        directory_url_ctors['s3'] = S3Url
        directory_url_ctors['file'] = FilesystemDirectoryUrl
    if len(file_url_ctors) == 0:
        file_url_ctors['s3'] = S3Url
        file_url_ctors['file'] = FilesystemFileUrl

        file_url_ctors['http'] = HttpFileUrl
        file_url_ctors['https'] = HttpFileUrl


def FileUrl(url: str, **kwargs: Any) -> BaseFileUrl:
    init_urls()
    parsed_url = urlparse(url)
    if parsed_url.scheme in file_url_ctors:
        ctor = file_url_ctors[parsed_url.scheme]

        out = ctor(url, **kwargs)
        if not isinstance(out, BaseFileUrl):
            raise TypeError(f"Not a file url: {url}")
        return out
    else:
        raise NotImplementedError(f"Teach me how to create FileUrls for {parsed_url.scheme}")


def DirectoryUrl(url: str, **kwargs: Any) -> BaseDirectoryUrl:
    init_urls()
    parsed_url = urlparse(url)
    if parsed_url.scheme in directory_url_ctors:
        ctor = directory_url_ctors[parsed_url.scheme]
        out = ctor(url, **kwargs)
        if not isinstance(out, BaseDirectoryUrl):
            raise TypeError(f"Not a directory url: {url}")
        return out
    else:
        raise NotImplementedError(f"Teach me how to create DirectoryUrls for {parsed_url.scheme}")



from sqlalchemy.engine import Engine, Connection
import sqlalchemy_redshift.dialect
from typing import Optional, Union


def _dialect(db: Optional[Union[Connection, Engine]]):
    if db is None:
        # use a common one which is likely to be compatible
        return sqlalchemy_redshift.dialect.RedshiftDialect()
    else:
        return db.dialect


def quote_schema_and_table(db: Optional[Union[Connection, Engine]],
                           schema: str,
                           table: str) -> str:
    """
    Prevent SQL injection when we're not able to use bind
    variables (e.g., passing a table name in SQL).

    http://bobby-tables.com
    http://bobby-tables.com/about

    e.g., for Vertica:

    bobby: Robert'); DROP TABLE Students;--
    quoted_bobby: "Robert'); DROP TABLE Students;--"

    """
    dialect = _dialect(db)
    return (dialect.preparer(dialect).quote(schema) + '.' +
            dialect.preparer(dialect).quote(table))


def quote_table_only(db: Optional[Union[Connection, Engine]], table: str) -> str:
    dialect = _dialect(db)
    return dialect.preparer(dialect).quote(table)


def quote_column_name(db: Optional[Union[Connection, Engine]], column_name: str) -> str:
    dialect = _dialect(db)
    return dialect.preparer(dialect).quote(column_name)


def quote_value(db: Optional[Union[Connection, Engine]], value: str) -> str:
    """
    Prevent SQL injection on literal string values in places when we're
    not able to use bind variables (e.g., using weird DB-specific
    commands/statements).

    http://bobby-tables.com
    http://bobby-tables.com/about

    e.g., for Vertica:

    bobby: Robert'); DROP TABLE Students;--
    quoted_bobby: 'Robert''); DROP TABLE Students;--'

    """
    dialect = _dialect(db)
    return dialect.preparer(dialect, initial_quote="'").quote(value)


def quote_user_name(db: Optional[Union[Connection, Engine]], user_name: str) -> str:
    dialect = _dialect(db)
    return dialect.preparer(dialect).quote_identifier(user_name)


def quote_group_name(db: Optional[Union[Connection, Engine]], group_name: str) -> str:
    dialect = _dialect(db)
    return dialect.preparer(dialect).quote_identifier(group_name)



from ..driver import DBDriver
import sqlalchemy
import vertica_python
import tempfile
import pathlib
from sqlalchemy.sql import text
from records_mover.db.quoting import quote_schema_and_table
from ...records.records_directory import RecordsDirectory
from ...records.records_format import BaseRecordsFormat
from contextlib import contextmanager
from sqlalchemy.schema import Table, Column
import logging
from typing import Iterator, Optional, IO, Union, List, Tuple, Type
from ...url.resolver import UrlResolver
from ...url.base import BaseDirectoryUrl
from ...records.unload_plan import RecordsUnloadPlan
from ...records.load_plan import RecordsLoadPlan
from .loader import VerticaLoader
from .unloader import VerticaUnloader
from ...utils.limits import (INT64_MIN, INT64_MAX,
                             FLOAT64_SIGNIFICAND_BITS,
                             num_digits)


logger = logging.getLogger(__name__)


class VerticaDBDriver(DBDriver):
    def __init__(self,
                 db: Union[sqlalchemy.engine.Connection, sqlalchemy.engine.Engine],
                 url_resolver: UrlResolver,
                 s3_temp_base_loc: Optional[BaseDirectoryUrl]=None) -> None:
        super().__init__(db)
        self._vertica_loader = VerticaLoader(url_resolver=url_resolver, db=self.db)
        self._vertica_unloader = VerticaUnloader(s3_temp_base_loc=s3_temp_base_loc, db=db)
        self.url_resolver = url_resolver

    def load(self,
             schema: str,
             table: str,
             load_plan: RecordsLoadPlan,
             directory: RecordsDirectory) -> None:
        self._vertica_loader.load(schema=schema,
                                  table=table,
                                  load_plan=load_plan,
                                  directory=directory)

    def can_load_this_format(self, source_records_format: BaseRecordsFormat) -> bool:
        return self._vertica_loader.can_load_this_format(source_records_format)

    def known_supported_records_formats_for_load(self) -> List[BaseRecordsFormat]:
        return self._vertica_loader.known_supported_records_formats_for_load()

    def load_from_fileobj(self,
                          schema: str,
                          table: str,
                          load_plan: RecordsLoadPlan,
                          fileobj: IO[bytes]) -> None:
        self._vertica_loader.load_from_fileobj(schema=schema,
                                               table=table,
                                               load_plan=load_plan,
                                               fileobj=fileobj)

    def unload(self,
               schema: str,
               table: str,
               unload_plan: RecordsUnloadPlan,
               directory: RecordsDirectory) -> int:
        return self._vertica_unloader.unload(schema=schema,
                                             table=table,
                                             unload_plan=unload_plan,
                                             directory=directory)

    def can_unload_this_format(self, target_records_format: BaseRecordsFormat) -> bool:
        return self._vertica_unloader.can_unload_this_format(target_records_format)

    def known_supported_records_formats_for_unload(self) -> List[BaseRecordsFormat]:
        return self._vertica_unloader.known_supported_records_formats_for_unload()

    def has_table(self, schema: str, table: str) -> bool:
        try:
            sql = f"SELECT 1 from {quote_schema_and_table(self.db, schema, table)} limit 0;"
            self.db.execute(sql)
            return True
        except sqlalchemy.exc.ProgrammingError:
            return False

    @contextmanager
    def temporary_loadable_directory_loc(self) -> Iterator[BaseDirectoryUrl]:
        with tempfile.TemporaryDirectory(prefix='vertica_load_location') as tempdir:
            output_url = pathlib.Path(tempdir).resolve().as_uri() + '/'
            yield self.url_resolver.directory_url(output_url)

    def best_scheme_to_load_from(self) -> str:
        return 'file'

    def schema_sql(self, schema: str, table: str) -> str:
        sql = text("SELECT EXPORT_OBJECTS('', :schema_and_table, false)")
        result = self.db.execute(sql, schema_and_table=f"{schema}.{table}").fetchall()

        if len(result) == 1:
            return result[0].EXPORT_OBJECTS
        else:
            # maybe a permission error?
            return super().schema_sql(schema, table)

    def best_records_format_variant(self,
                                    records_format_type: str) -> Optional[str]:
        if records_format_type == 'delimited':
            return 'vertica'
        return None

    def can_load_from_fileobjs(self) -> bool:
        return True

    def load_failure_exception(self) -> Type[Exception]:
        return vertica_python.errors.CopyRejected

    def table(self,
              schema: str,
              table: str) -> Table:
        # sqlalchemy-vertica driver uses system tables that are
        # suuuuper slow in Vertica databases--this is a compromise
        # that pulls column info but not other things
        #
        # https://docs.sqlalchemy.org/en/latest/core/metadata.html#sqlalchemy.schema.Column
        # https://docs.sqlalchemy.org/en/latest/core/reflection.html#
        #    sqlalchemy.engine.reflection.Inspector.get_columns
        columns = [Column(colinfo['name'],
                          type_=colinfo['type'],
                          nullable=colinfo['nullable'],
                          default=colinfo['default'],
                          **colinfo.get('attrs', {}))
                   for colinfo in self.db_engine.dialect.get_columns(self.db_engine,
                                                                     table,
                                                                     schema=schema)]
        t = Table(table, self.meta, schema=schema, *columns)
        return t

    def integer_limits(self,
                       type_: sqlalchemy.types.Integer) ->\
            Optional[Tuple[int, int]]:
        if isinstance(type_, sqlalchemy.sql.sqltypes.INTEGER):
            return (INT64_MIN, INT64_MAX)
        return super().integer_limits(type_)

    def fp_constraints(self,
                       type_: sqlalchemy.types.Float) ->\
            Optional[Tuple[int, int]]:
        if isinstance(type_, sqlalchemy.sql.sqltypes.FLOAT):
            return (64, FLOAT64_SIGNIFICAND_BITS)
        return super().fp_constraints(type_)

    def type_for_integer(self,
                         min_value: Optional[int],
                         max_value: Optional[int]) -> sqlalchemy.types.TypeEngine:
        """Find correct integral column type to fit the given min and max integer values"""

        if min_value is not None and max_value is not None:
            if min_value >= INT64_MIN and max_value <= INT64_MAX:
                return sqlalchemy.sql.sqltypes.INTEGER()
            else:
                num_digits_min = num_digits(min_value)
                num_digits_max = num_digits(max_value)
                digit_count = max(num_digits_min, num_digits_max)
                return self.type_for_fixed_point(precision=digit_count,
                                                 scale=0)
        return super().type_for_integer(min_value, max_value)

    def type_for_floating_point(self,
                                fp_total_bits: int,
                                fp_significand_bits: int) -> sqlalchemy.sql.sqltypes.Numeric:
        if fp_significand_bits > FLOAT64_SIGNIFICAND_BITS:
            logger.warning("Falling back to Vertica 64-bit FLOAT type, as Vertica doesn't support "
                           f"fp_significand_bits>=54 (requested: {fp_significand_bits}")
            return sqlalchemy.sql.sqltypes.Float(precision=FLOAT64_SIGNIFICAND_BITS)
        return super().type_for_floating_point(fp_total_bits=fp_total_bits,
                                               fp_significand_bits=fp_significand_bits)



from typing import Iterable, Callable, Any, Optional, Dict, List
import docstring_parser
import enum
import inspect
import typing
from typing_inspect import get_origin, is_callable_type, get_args, is_union_type
import collections
import logging
from collections import OrderedDict
from ..types import JsonSchema
from .json_schema_document import JsonSchemaDocument, DefaultValue
from .json_schema_array_document import JsonSchemaArrayDocument
from .json_parameter import JsonParameter


logger = logging.getLogger(__name__)


PythonType = Any


def is_iterable_type(python_type: PythonType) -> bool:
    return get_origin(python_type) in [
        collections.abc.Iterable,
        typing.Iterable,
        typing.List,
        list
    ]


# copied in until a release is cut:
#  https://github.com/ilevkivskyi/typing_inspect/blob/master/typing_inspect.py#L119
def is_optional_type(tp: PythonType) -> bool:
    """Returns `True` if the type is `type(None)`, or is a direct `Union` to `type(None)`, such as `Optional[T]`.  # noqa
    NOTE: this method inspects nested `Union` arguments but not `TypeVar` definitions (`bound`/`constraint`). So it
    will return `False` if
     - `tp` is a `TypeVar` bound, or constrained to, an optional type
     - `tp` is a `Union` to a `TypeVar` bound or constrained to an optional type,
     - `tp` refers to a *nested* `Union` containing an optional type or one of the above.
    Users wishing to check for optionality in types relying on type variables might wish to use this method in
    combination with `get_constraints` and `get_bound`
    """

    if tp is type(None):  # noqa
        return True
    elif is_union_type(tp):
        return any(is_optional_type(tt) for tt in get_args(tp, evaluate=True))
    else:
        return False


def is_impractical_type(python_type: PythonType) -> bool:
    # can't write code and pass it around in JSON!
    return (is_callable_type(python_type) or
            type(python_type) == enum.EnumMeta)


def parse_python_parameter_type(name: str,
                                python_type: PythonType,
                                description: Optional[str],
                                optional: bool,
                                default: DefaultValue) ->\
                Optional[JsonParameter]:
    if python_type == str:
        return JsonParameter(name, JsonSchemaDocument(json_type='string',
                                                      default=default,
                                                      description=description), optional=optional)
    elif python_type == int:
        return JsonParameter(name, JsonSchemaDocument(json_type='integer',
                                                      default=default,
                                                      description=description), optional=optional)
    elif python_type == float:
        return JsonParameter(name, JsonSchemaDocument(json_type='number',
                                                      default=default,
                                                      description=description), optional=optional)
    elif python_type == bool:
        return JsonParameter(name, JsonSchemaDocument(json_type='boolean',
                                                      default=default,
                                                      description=description), optional=optional)
    elif is_optional_type(python_type):
        inner_type = get_args(python_type, evaluate=True)[0]
        if not is_impractical_type(inner_type):
            if default is None:
                default = inspect.Parameter.empty
            return parse_python_parameter_type(name,
                                               inner_type,
                                               description=description,
                                               optional=True,
                                               default=default)
    elif is_iterable_type(python_type):
        inner_type = get_args(python_type, evaluate=True)[0]
        inner_results = parse_python_parameter_type(name, inner_type,
                                                    description=None,
                                                    optional=optional,
                                                    default=inspect.Parameter.empty)

        if inner_results is not None:
            return JsonParameter(inner_results.name,
                                 JsonSchemaArrayDocument(json_type='array',
                                                         items=inner_results.json_schema_document,
                                                         description=description,
                                                         default=default),
                                 optional=inner_results.optional)
    else:
        raise NotImplementedError(f"Teach me how to handle Python type {python_type} "
                                  f"(parameter: {name})")
    return None


def method_signature_to_json_schema(m: Callable[..., Any],
                                    special_handling: Dict[str, List[JsonParameter]],
                                    parameters_to_ignore: Iterable[str]) -> JsonSchema:
    """Creates a JSON schema representing the parameters to a method"""
    signature = inspect.signature(m)
    generated: JsonSchema = {
        'type': 'object',
        'properties': {}
    }

    properties: Dict[str, JsonSchemaDocument] = OrderedDict()
    required: List[str] = []
    method_docs = docstring_parser.parse(m.__doc__)

    def parse_parameter(param_name: str, python_type: PythonType, default: DefaultValue) -> None:
        if param_name in parameters_to_ignore:
            return
        try:
            param_docs = next(param
                              for param in method_docs.params
                              if param.arg_name == param_name)
            description = param_docs.description
        except StopIteration:
            print(f"Could not find docs for {param_name}")
            raise

        if param_name in special_handling:
            for substitute_parameter in special_handling[param_name]:
                properties[substitute_parameter.name] = substitute_parameter.json_schema_document
                if not substitute_parameter.optional:
                    required.append(substitute_parameter.name)
                if properties[substitute_parameter.name].description is None:
                    properties[substitute_parameter.name].description = description
        else:
            optional = default != inspect.Parameter.empty
            results = parse_python_parameter_type(name=param_name,
                                                  python_type=python_type,
                                                  description=description,
                                                  optional=optional,
                                                  default=default)

            if results is not None:
                properties[results.name] = results.json_schema_document
                if not results.optional:
                    required.append(results.name)

    for param_name, parameter in signature.parameters.items():
        python_type = parameter.annotation
        default = parameter.default
        parse_parameter(param_name, python_type, default)
    generated['properties'] = OrderedDict()
    for name, prop in properties.items():
        generated['properties'][name] = prop.to_data()
    generated['required'] = required
    return generated



import json
import numpy as np
from ..utils.structures import map_keys, snake_to_camel, nest_dict
from pandas import DataFrame
from typing import List, Dict, Any


# http://stackoverflow.com/questions/27050108/convert-numpy-type-to-python
class NumPyJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(NumPyJSONEncoder, self).default(obj)


def dataframe_to_nested_dicts(df: DataFrame,
                              to_camel: bool=False) -> List[Dict[str, Any]]:
    """
    This turns database results (expressed as a pandas dataframe) into
    potentially-nested dicts. It uses a '.' in the column names
    as hints to nest.

    e.g., the dataframe created from this query result:

    +-----+---------+--------------+
    | abc | foo.bar | foo.baz.bing |
    +-----+---------+--------------+
    |  1  | 5       | 'bazzle'     |
    +-----+---------+--------------+

    results in this dict:

    {'abc': 1, 'foo': {'bar': 5, 'baz': {'bing': 'bazzle'}}}
    """
    # 'records' output is like:
    # [{"col1": 123, "col2": "abc"}, {"col1": 456, "col2": "xyz"}]
    records = df.to_dict(orient='records')
    if to_camel:
        records = map(lambda d: map_keys(snake_to_camel, d), records)
    records = map(nest_dict, records)
    return list(records)


def json_dumps(item: str) -> Any:
    return json.dumps(item, cls=NumPyJSONEncoder)


def purge_unnamed_unused_columns(df: DataFrame) -> DataFrame:
    # Some spreadsheets will drop a trailing ',' after the
    # last column in the header with actual content when
    # exporting to CSV, leaving you with an empty column.
    # Pandas helpfully names this empty column for you as
    # "unnamed: 1", or maybe "Unnamed: 1" (not sure why/when
    # that differs).  Let's clean those up.
    for column in df:
        if column.startswith('Unnamed: ') or column.startswith('unnamed: '):
            if not df[column].notnull().any():
                df = df.drop(column, axis=1)
    return df
